# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import tensorflow as tf


def convert_sync_batchnorm(model: tf.keras.Model) -> tf.keras.Model:
    """Replace BatchNormalization layers to SyncBatchNormalization in place.
    WARNINGS:
    * This function is tested only with efficientnet and resnet
    * The returned model has shared layers with the input one. One of them should be disposed.
    """

    layer2newtensor = {}
    in_ = tf.keras.layers.Input(model.input.shape[1:])
    layer2newtensor[model.input.name] = in_
    for layer in model.layers[1:]:
        assert len(layer.inbound_nodes) == 1
        if isinstance(layer.inbound_nodes[0].inbound_layers, list):  # mutliple inputs
            x = [layer2newtensor[in_l.name] for in_l in layer.inbound_nodes[0].inbound_layers]
        else:
            x = layer2newtensor[layer.inbound_nodes[0].inbound_layers.name]
        if isinstance(layer, tf.keras.layers.BatchNormalization):
            layer = tf.keras.layers.experimental.SyncBatchNormalization(**layer.get_config())

        if "truediv" in layer.name:
            # efficeientnet edge case
            # https://github.com/keras-team/keras/blob/v2.9.0/keras/applications/efficientnet.py#L334
            x = layer(x, layer.inbound_nodes[0]._flat_arguments[1])
        else:
            x = layer(x)

        layer2newtensor[layer.name] = x
    out_ = layer2newtensor[model.layers[-1].name]
    new_model = tf.keras.Model(inputs=in_, outputs=out_, name=model.name)
    new_model.set_weights(model.get_weights())
    return new_model



from __future__ import annotations

from collections.abc import Callable, Sequence

import tensorflow as tf


def filter_classes(
    ds: tf.data.Dataset,
    class_list: Sequence[int] | None = None,
    y_parser: Callable = lambda y: y,
) -> tf.data.Dataset:
    """
    Filters a dataset by class id.

    Args:
        ds: A `tf.data.Dataset` object.
        class_list: An optional `Sequence` of integers representing the classes
            to include in the dataset. If `None`, all classes are included.
        y_parser: A callable function used to parse the class id from the y outputs.

    Returns:
        A `tf.data.Dataset` object filtered by class id.
    """

    if class_list is not None:
        class_list = tf.constant(class_list)
        ds = ds.filter(lambda x, y, *args: tf.reduce_any(tf.equal(y_parser(y), class_list)))

    return ds


def create_grouped_dataset(
    ds: tf.data.Dataset,
    window_size: int,
    total_examples: int | None = None,
    buffer_size: int | None = None,
    y_parser: Callable = lambda y: y,
) -> list[tf.data.Dataset]:
    """
    Creates a list of datasets grouped by class id.

    Args:
        ds: A `tf.data.Dataset` object.
        window_size: An integer representing the dataset cardinality.
        total_examples: An integer representing the maximum number of examples
            to include in the dataset. If `None`, all examples are included.
        buffer_size: An optional integer representing the size of the buffer
            for shuffling. Default is None.
        y_parser: A callable function used to parse the class id from the y outputs.

    Returns:
        A List of `tf.data.Dataset` objects grouped by class id.
    """
    # NOTE: We need to cast the key_func as the group_op expects an int64.
    grouped_by_cid = ds.group_by_window(
        key_func=lambda x, y, *args: tf.cast(y_parser(y), dtype=tf.int64),
        reduce_func=lambda key, ds: ds.batch(window_size),
        window_size=window_size,
    )

    cid_datasets = []
    for elem in grouped_by_cid:
        cid_ds = tf.data.Dataset.from_tensor_slices(elem)
        if total_examples is not None:
            cid_ds = cid_ds.take(total_examples)
        if buffer_size is not None:
            cid_ds = cid_ds.shuffle(buffer_size)
        cid_datasets.append(cid_ds.repeat())

    return cid_datasets


def create_choices_dataset(num_classes: int, examples_per_class: int) -> tf.data.Dataset:
    """
    Creates a dataset that generates random integers between 0 and `num_classes`.
    Integers will be generated in contiguous blocks of size `examples_per_class`.
    Integers are sampled without replacement and are not selected again until all
    other interger values have been sampled.

    Args:
        num_classes: An integer representing the total number of classes in the dataset.
        examples_per_class: An integer representing the number of examples per class.

    Returns:
        A `tf.data.Dataset` object representing the dataset with random choices.
    """
    return (
        tf.data.Dataset.range(num_classes)
        .shuffle(num_classes)
        .map(lambda z: [[z] * examples_per_class], name="repeat_cid")
        .flat_map(tf.data.Dataset.from_tensor_slices)
        .repeat()
    )


def apply_augmenter_ds(ds: tf.data.Dataset, augmenter: Callable, warmup: int = 0) -> tf.data.Dataset:
    """
    Applies an augmenter function to a dataset batch and optionally delays
    applying the function for `warmup` number of examples.

    Args:
        ds: A `tf.data.Dataset` object.
        augmenter: A callable function used to apply data augmentation to
            individual examples within each batch. If `None`, no data
            augmentation is applied.
        warmup: An integer representing the number of examples to wait
            before applying the data augmentation function.

    Returns:
        A `tf.data.Dataset` object with the applied augmenter.
    """
    if not warmup:
        return ds.map(augmenter, name="augmenter")

    aug_ds = ds.map(augmenter, name="augmenter").skip(warmup)

    try:
        count_ds = tf.data.Dataset.counter()
    except AttributeError:
        count_ds = tf.data.experimental.Counter()

    ds = tf.data.Dataset.choose_from_datasets(
        [ds.take(warmup), aug_ds],
        count_ds.map(lambda x: tf.cast(0, dtype=tf.int64) if x < warmup else tf.cast(1, dtype=tf.int64)),
    )

    return ds


def TFDataSampler(
    ds: tf.data.Dataset,
    classes_per_batch: int = 2,
    examples_per_class_per_batch: int = 2,
    class_list: Sequence[int] | None = None,
    total_examples_per_class: int | None = None,
    augmenter: Callable | None = None,
    load_fn: Callable | None = None,
    warmup: int = 0,
    label_output: int | str | None = None,
) -> tf.data.Dataset:
    """
    Returns a `tf.data.Dataset` object that generates batches of examples with
    equal number of examples per class. The input dataset cardinality must be
    finite and known.

    Args:
        ds: A `tf.data.Dataset` object representing the original dataset.
        classes_per_batch: An integer specifying the number of classes per batch.
        examples_per_class_per_batch: An integer specifying the number of examples
            per class per batch.
        class_list: An optional sequence of integers representing the class IDs
            to include in the dataset. If `None`, all classes in the original
            dataset will be used.
        total_examples_per_class: An optional integer representing the total
            number of examples per class to use.  If `None`, all examples for
            each class will be used.
        augmenter: An optional function to apply data augmentation to each
            example in a batch.
        load_fn: An optional callable function for loading real examples from `x`.
            It is useful for loading images from their corresponding file path
            provided in `x` or similar situations.
        warmup: An integer specifying the number of examples to use for unaugmented
            warmup.
        label_output: An optional integer or string representing the label output
            used to create the balanced dataset batches. If `None`, y is assumed
            to be a 1D integer tensor containing the class IDs. If `int`, y is
            assumed to be a tuple of tensors with the class IDs in the element
            specified by `label_output`. If `str`, y is assumed to be a dictionary
            with the class IDs in the key specified by `label_output`.
    Returnsk:
        A `tf.data.Dataset` object representing the balanced dataset for few-shot learning tasks.

    """
    if ds.cardinality() == tf.data.INFINITE_CARDINALITY:
        raise ValueError("Dataset must be finite.")
    if ds.cardinality() == tf.data.UNKNOWN_CARDINALITY:
        raise ValueError("Dataset cardinality must be known.")

    def y_parser(y):
        return y if label_output is None else y[label_output]

    window_size = ds.cardinality().numpy()
    batch_size = examples_per_class_per_batch * classes_per_batch

    ds = filter_classes(ds, class_list=class_list, y_parser=y_parser)
    ds = create_grouped_dataset(
        ds,
        window_size=window_size,
        total_examples=total_examples_per_class,
        y_parser=y_parser,
    )
    choices_ds = create_choices_dataset(
        len(ds),
        examples_per_class=examples_per_class_per_batch,
    )
    ds = tf.data.Dataset.choose_from_datasets(ds, choices_ds)

    if load_fn is not None:
        ds = ds.map(load_fn, name="load_fn")

    if augmenter is not None:
        ds = apply_augmenter_ds(ds, augmenter, warmup)

    ds = ds.repeat().batch(batch_size).prefetch(tf.data.AUTOTUNE)

    return ds



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import TypeVar

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor, IntTensor

from .memory_samplers import MultiShotMemorySampler
from .samplers import Augmenter

T = TypeVar("T", FloatTensor, IntTensor)


def load_image(path: str, target_size: tuple[int, int] | None = None) -> T:
    image_string = tf.io.read_file(path)
    image: T = tf.image.decode_jpeg(image_string, channels=3)
    image = tf.image.convert_image_dtype(image, tf.float32)
    if target_size:
        image = tf.image.resize(image, target_size, method=tf.image.ResizeMethod.LANCZOS3)
        image = tf.clip_by_value(image, 0.0, 1.0)
    return image


class MultiShotFileSampler(MultiShotMemorySampler):
    def __init__(
        self,
        x,
        y,
        load_example_fn: Callable = load_image,
        classes_per_batch: int = 2,
        examples_per_class_per_batch: int = 2,
        steps_per_epoch: int = 1000,
        class_list: Sequence[int] | None = None,
        total_examples_per_class: int | None = None,
        augmenter: Augmenter | None = None,
        warmup: int = -1,
    ):
        super().__init__(
            x,
            y,
            load_example_fn=load_example_fn,
            classes_per_batch=classes_per_batch,
            examples_per_class_per_batch=examples_per_class_per_batch,
            steps_per_epoch=steps_per_epoch,
            class_list=class_list,
            total_examples_per_class=total_examples_per_class,
            augmenter=augmenter,
            warmup=warmup,
        )



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Specialized callbacks that track similarity metrics during training"""
from __future__ import annotations

import math
from collections.abc import MutableMapping, Sequence
from pathlib import Path

import numpy as np
import tensorflow as tf
from tensorflow.keras.callbacks import Callback

from .classification_metrics import ClassificationMetric, make_classification_metric
from .evaluators import Evaluator, MemoryEvaluator
from .matchers import ClassificationMatch
from .models import SimilarityModel
from .retrieval_metrics import RetrievalMetric
from .types import FloatTensor, IntTensor, Tensor
from .utils import unpack_lookup_distances, unpack_lookup_labels, unpack_results


class EvalCallback(Callback):
    """Epoch end evaluation callback that build a test index and evaluate
    model performance on it.

    This evaluation only run at epoch_end as it is computationally very
    expensive.

    """

    def __init__(
        self,
        queries: Tensor,
        query_labels: Sequence[int],
        targets: Tensor,
        target_labels: Sequence[int],
        distance: str = "cosine",
        metrics: Sequence[str | ClassificationMetric] = ["binary_accuracy"],
        tb_logdir: str | None = None,
        k: int = 1,
        matcher: str | ClassificationMatch = "match_nearest",
        distance_thresholds: FloatTensor | None = None,
        known_classes: IntTensor | None = None,
        retrieval_metrics: Sequence[RetrievalMetric] | None = None,
    ):
        """Evaluate model matching quality against a validation dataset at
        epoch end.

        Args:
            queries: Test examples that will be tested against the built index.

            query_labels: Queries nearest neighbors expected labels.

            targets: Examples that are indexed.

            target_labels: Target examples labels.

            distance: Distance function used to compute pairwise distance
            between examples embeddings.

            metrics: List of
            'tf.similarity.classification_metrics.ClassificationMetric()` to
            compute during the evaluation. Defaults to ['binary_accuracy',
            'f1score'].

            tb_logdir: Where to write TensorBoard logs. Defaults to None.

            k: The number of nearest neighbors to return for each query.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            distance_thresholds: A 1D tensor denoting the distances points at
            which we compute the metrics. If None, distance_thresholds is set
            to tf.constant([math.inf])

            known_classes: Optional. If specified the validation set will be
            split into two subsets based on whether a class was seen or not
            during training. This should be a list of classes seen during
            training.

        """
        super().__init__()
        t_query_labels = tf.convert_to_tensor(np.array(query_labels))
        self.target_labels = tf.convert_to_tensor(np.array(target_labels))

        # Ensure all label tensors have the same dtype.
        self.target_labels = tf.cast(self.target_labels, dtype=t_query_labels.dtype)

        self.targets = targets
        self.distance = distance
        self.evaluator = MemoryEvaluator()
        # typing requires this weird formulation of creating a new list
        self.classification_metrics: list[ClassificationMetric] = [make_classification_metric(m) for m in metrics]
        self.retrieval_metrics = retrieval_metrics if retrieval_metrics is not None else []
        self.k = k
        self.matcher = matcher

        if distance_thresholds is not None:
            self.distance_thresholds = distance_thresholds
        else:
            self.distance_thresholds = tf.constant([math.inf])

        if known_classes is not None:
            known_classes = tf.cast(known_classes, dtype=t_query_labels.dtype)
            self.split_validation = True
        else:
            self.split_validation = False

        if tb_logdir:
            tb_logdir = str(Path(tb_logdir) / "index/")
            self.tb_writer = tf.summary.create_file_writer(tb_logdir)
            print("TensorBoard logging enable in %s" % tb_logdir)
        else:
            self.tb_writer = None

        # Use broadcasting to do a y X known_classes equality check. By adding
        # a dim to the start of known_classes and a dim to the end of y, this
        # essentially checks `for ck in known_classes: for cy in y: ck == cy`.
        # We then reduce_any to find all rows in y that match at least one
        # class in known_classes.
        # See https://numpy.org/doc/stable/user/basics.broadcasting.html
        if self.split_validation:
            known_classes = tf.reshape(known_classes, (-1))
            broadcast_classes = tf.expand_dims(known_classes, axis=0)
            broadcast_labels = tf.expand_dims(t_query_labels, axis=-1)
            known_mask = tf.math.reduce_any(broadcast_classes == broadcast_labels, axis=1)
            known_idxs = tf.squeeze(tf.where(known_mask))
            unknown_idxs = tf.squeeze(tf.where(~known_mask))

        with tf.device("/cpu:0"):
            if self.split_validation:
                self.queries_known = tf.gather(queries, indices=known_idxs)
                self.query_labels_known = tf.gather(t_query_labels, indices=known_idxs)
                # Expand to 2D if we only have a single example
                if tf.rank(self.queries_known) == 1:
                    self.queries_known = tf.expand_dims(self.queries_known, axis=0)
                    self.query_labels_known = tf.expand_dims(self.query_labels_known, axis=0)

                self.queries_unknown = tf.gather(queries, indices=unknown_idxs)
                self.query_labels_unknown = tf.gather(t_query_labels, indices=unknown_idxs)
                # Expand to 2D if we only have a single example
                if tf.rank(self.queries_unknown) == 1:
                    self.queries_unknown = tf.expand_dims(self.queries_unknown, axis=0)
                    self.query_labels_unknown = tf.expand_dims(self.query_labels_unknown, axis=0)
            else:
                self.queries_known = queries
                self.query_labels_known = t_query_labels
                self.queries_unknown = None
                self.query_labels_unknown = None

    def on_epoch_end(self, epoch: int, logs: MutableMapping | None = None):
        """Computes the eval metrics at the end of each epoch.

        NOTE: This method resets the index and batch adds the target embeddings
        to the index using the new embeddings generated by the current version
        of the model.
        """
        _ = epoch
        if logs is None:
            logs = {}

        # reset the index
        self.model.reset_index()

        # rebuild the index
        self.model.index(self.targets, self.target_labels, verbose=0)

        known_results = _compute_metrics(
            queries=self.queries_known,
            query_labels=self.query_labels_known,
            model=self.model,
            evaluator=self.evaluator,
            classification_metrics=self.classification_metrics,
            retrieval_metrics=self.retrieval_metrics,
            k=self.k,
            matcher=self.matcher,
            distance_thresholds=self.distance_thresholds,
        )

        mstr = []

        if self.split_validation:
            unknown_results = _compute_metrics(
                queries=self.queries_unknown,
                query_labels=self.query_labels_unknown,
                model=self.model,
                evaluator=self.evaluator,
                classification_metrics=self.classification_metrics,
                retrieval_metrics=self.retrieval_metrics,
                k=self.k,
                matcher=self.matcher,
                distance_thresholds=self.distance_thresholds,
            )

            for a, b in zip(known_results, unknown_results):
                mstr.extend(
                    unpack_results(
                        a,
                        epoch=epoch,
                        logs=logs,
                        tb_writer=self.tb_writer,
                        name_suffix="_known_classes",
                    )
                )
                mstr.extend(
                    unpack_results(
                        b,
                        epoch=epoch,
                        logs=logs,
                        tb_writer=self.tb_writer,
                        name_suffix="_unknown_classes",
                    )
                )
        else:
            for a in known_results:
                mstr.extend(
                    unpack_results(
                        a,
                        epoch=epoch,
                        logs=logs,
                        tb_writer=self.tb_writer,
                    )
                )
        self.model.reset_index()


def _compute_metrics(
    queries: Tensor,
    query_labels: IntTensor,
    model: SimilarityModel,
    evaluator: Evaluator,
    classification_metrics: Sequence[ClassificationMetric],
    retrieval_metrics: Sequence[RetrievalMetric],
    k: int,
    matcher: str | ClassificationMatch,
    distance_thresholds: FloatTensor,
) -> tuple[dict[str, np.ndarray], dict[str, np.ndarray]]:
    """Compute the classification metrics.

    Args:
        queries: A Tensor of embeddings representing the queries.

        query_labels: An IntTensor representing the class ids associated with
            the queries.

        model: The current similarity model.

        evaluator: An Evalutar object for evaluating the index performance.

        metrics: A list of classification metrics objects.

        k: The number of nearest neighbors to return for each query.

        matcher: {'match_nearest', 'match_majority_vote'} or ClassificationMatch
        object. Defines the classification matching, e.g., match_nearest will
        count a True Positive if the query_label is equal to the label of the
        nearest neighbor and the distance is less than or equal to the distance
        threshold.

        distance_thresholds: A 1D tensor denoting the distances points at which
        we compute the metrics.

    Returns:
        A Python dict mapping the metric name to the copmuted value.
    """
    lookups = model.lookup(queries, k=k, verbose=0)
    lookup_distances = unpack_lookup_distances(lookups, model.dtype)
    lookup_labels = unpack_lookup_labels(lookups, query_labels.dtype)

    if classification_metrics:
        classification_results = evaluator.evaluate_classification(
            query_labels=query_labels,
            lookup_labels=lookup_labels,
            lookup_distances=lookup_distances,
            distance_thresholds=distance_thresholds,
            metrics=classification_metrics,
            matcher=matcher,
            verbose=0,
        )
        # The callbacks don't set a distance theshold so we remove it here.
        classification_results.pop("distance")
    else:
        classification_results = {}

    if retrieval_metrics:
        retrieval_results = evaluator.evaluate_retrieval(
            target_labels=query_labels,
            lookups=lookups,
            retrieval_metrics=retrieval_metrics,
        )
    else:
        retrieval_results = {}

    return classification_results, retrieval_results



import re

import pytest

from tensorflow_similarity.architectures import resnet18


def test_include_top():
    input_shape = (32, 32, 3)
    resnet = resnet18.ResNet18Sim(input_shape, include_top=True)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-2].name == "gem_pool"
    assert resnet.layers[-2].p == 3.0
    # The default is l2_norm True, so we expect the last layer to be
    # MetricEmbedding.
    assert re.match("metric_embedding", resnet.layers[-1].name) is not None


def test_l2_norm_false():
    input_shape = (32, 32, 3)
    resnet = resnet18.ResNet18Sim(input_shape, include_top=True, l2_norm=False)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-2].name == "gem_pool"
    assert resnet.layers[-2].p == 3.0
    # If l2_norm is False, we should return a dense layer as the last layer.
    assert re.match("dense", resnet.layers[-1].name) is not None


@pytest.mark.parametrize(
    "pooling, name", zip(["gem", "avg", "max"], ["gem_pool", "avg_pool", "max_pool"]), ids=["gem", "avg", "max"]
)
def test_include_top_false(pooling, name):
    input_shape = (32, 32, 3)
    resnet = resnet18.ResNet18Sim(input_shape, include_top=False, pooling=pooling)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-1].name == name



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

import numpy as np
import tensorflow as tf
from absl.testing import parameterized
from tensorflow.keras.optimizers import schedules
from tensorflow.python.framework import combinations

from tensorflow_similarity.schedules import WarmupCosineDecay


def _maybe_serialized(lr_decay, serialize_and_deserialize):
    if serialize_and_deserialize:
        serialized = schedules.serialize(lr_decay)
        return schedules.deserialize(serialized)
    else:
        return lr_decay


def np_warmup_cosine_decay(*, step, max_lr, total_steps, warmup_steps, alpha=0.0):
    step = min(step, total_steps)
    if step < warmup_steps:
        decayed = step / warmup_steps
    else:
        decay_steps = total_steps - warmup_steps
        completed_fraction = (step - warmup_steps) / decay_steps
        cosine_decay = 0.5 * (1.0 + math.cos(math.pi * completed_fraction))
        decayed = (1.0 - alpha) * cosine_decay + alpha
    return max_lr * decayed


@combinations.generate(
    combinations.combine(
        mode=["graph", "eager"],
        serialize=[False, True],
    )
)
class WarmupCosineDecayTest(tf.test.TestCase, parameterized.TestCase):
    def testDecay(self, serialize):
        num_training_steps = 1000
        warmup_steps = 500
        max_lr = 1.0
        for step in range(0, 1500, 250):
            decayed_lr = WarmupCosineDecay(max_lr, num_training_steps, warmup_steps)
            decayed_lr = _maybe_serialized(decayed_lr, serialize)
            expected = np_warmup_cosine_decay(
                step=step,
                max_lr=max_lr,
                total_steps=num_training_steps,
                warmup_steps=warmup_steps,
            )
            self.assertAllClose(self.evaluate(decayed_lr(step)), expected, 1e-6)

    def testAlpha(self, serialize):
        num_training_steps = 1000
        warmup_steps = 500
        max_lr = 1.0
        alpha = 0.1
        for step in range(0, 1500, 250):
            decayed_lr = WarmupCosineDecay(max_lr, num_training_steps, warmup_steps, alpha)
            decayed_lr = _maybe_serialized(decayed_lr, serialize)
            expected = np_warmup_cosine_decay(
                step=step,
                max_lr=max_lr,
                total_steps=num_training_steps,
                warmup_steps=warmup_steps,
                alpha=alpha,
            )
            self.assertAllClose(self.evaluate(decayed_lr(step)), expected, 1e-6)

    def testFloat64InitLearningRate(self, serialize):
        num_training_steps = 1000
        warmup_steps = 500
        max_lr = np.float64(1.0)
        for step in range(0, 1500, 250):
            decayed_lr = WarmupCosineDecay(max_lr, num_training_steps, warmup_steps)
            decayed_lr = _maybe_serialized(decayed_lr, serialize)
            expected = np_warmup_cosine_decay(
                step=step,
                max_lr=max_lr,
                total_steps=num_training_steps,
                warmup_steps=warmup_steps,
            )
            self.assertAllClose(self.evaluate(decayed_lr(step)), expected, 1e-6)

    def testMaxWarmupSteps(self, serialize):
        num_training_steps = 1000
        warmup_steps = 1000
        max_lr = 1.0
        msg = "warmup_steps must be less than the total steps"
        with self.assertRaisesRegex(ValueError, msg):
            _ = WarmupCosineDecay(max_lr, num_training_steps, warmup_steps)


if __name__ == "__main__":
    tf.test.main()



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""SimCLR Loss.
    A Simple Framework for Contrastive Learning of Visual Representations
    https://arxiv.org/abs/2002.05709
"""
from __future__ import annotations

from typing import Any

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor


@tf.keras.utils.register_keras_serializable(package="Similarity")
class SimCLRLoss(tf.keras.losses.Loss):
    """SimCLR Loss.
    [A Simple Framework for Contrastive Learning of Visual Representations](https://arxiv.org/abs/2002.05709)
    code adapted from [original github](https://github.com/google-research/simclr/tree/master/tf2)
    """

    LARGE_NUM = 1e9

    def __init__(self, temperature: float = 0.05, **kwargs):
        super().__init__(**kwargs)
        self.temperature = temperature

    def contrast(self, hidden1: FloatTensor, hidden2: FloatTensor) -> FloatTensor:

        # local replica batch size
        batch_size = tf.shape(hidden1)[0]

        if not tf.distribute.in_cross_replica_context():
            # SimCLR loss computes similarity with 2N-1 other examples, when N is the global_batch_size.
            # In distributed training, each replica sees only N/n_replicas examples and to compare with all 2N-1
            # examples we must first get them from other replicas.
            strategy = tf.distribute.get_strategy()
            hidden1_large = SimCLRLoss.cross_replica_concat(hidden1, strategy)
            hidden2_large = SimCLRLoss.cross_replica_concat(hidden2, strategy)
            enlarged_batch_size = tf.shape(hidden1_large)[0]
            replica_context = tf.distribute.get_replica_context()
            labels_idx = tf.range(batch_size) + replica_context.replica_id_in_sync_group * batch_size
            labels = tf.one_hot(labels_idx, enlarged_batch_size * 2)
            masks = tf.one_hot(labels_idx, enlarged_batch_size)
        else:
            hidden1_large = hidden1
            hidden2_large = hidden2
            labels = tf.one_hot(tf.range(batch_size), batch_size * 2)
            masks = tf.one_hot(tf.range(batch_size), batch_size)

        # compute pairwise
        ab = tf.matmul(hidden1, hidden2_large, transpose_b=True) / self.temperature
        aa = tf.matmul(hidden1, hidden1_large, transpose_b=True) / self.temperature

        # set the diagonal to a large negative number to ensure that z_i * z_i
        # is close to zero in the cross entropy.
        aa = aa - masks * SimCLRLoss.LARGE_NUM

        distances = tf.concat((ab, aa), axis=1)

        # 1D tensor
        per_example_loss: FloatTensor = tf.nn.softmax_cross_entropy_with_logits(labels, distances)

        return per_example_loss

    def call(self, za: FloatTensor, zb: FloatTensor) -> FloatTensor:
        """Compute the loss.

        Args:
            za: Embedding A
            zb: Embedding B

        Returns:
            loss: Per replica loss
        """
        # We expect za and zb to be rank 2 tensors.
        za = tf.math.l2_normalize(za, axis=1)
        zb = tf.math.l2_normalize(zb, axis=1)

        loss_a = self.contrast(za, zb)
        loss_b = self.contrast(zb, za)

        # When used inside of built-in training loops such as Model.fit, tf.keras.losses.Reduction
        # will default to SUM_OVER_BATCH_SIZE which takes into account the global batch size across all GPUs.
        loss: FloatTensor = loss_a + loss_b
        return loss

    @staticmethod
    def cross_replica_concat(tensor, strategy):
        """Reduce a concatenation of the `tensor` across TPU cores.
        Args:
            tensor: tensor to concatenate.
            strategy: A `tf.distribute.Strategy`. If not set, CPU execution is assumed.
        Returns:
            Tensor of the same rank as `tensor` with first dimension `num_replicas`
            times larger.
        """
        num_replicas = strategy.num_replicas_in_sync

        replica_context = tf.distribute.get_replica_context()
        with tf.name_scope("cross_replica_concat"):
            # This creates a tensor that is like the input tensor but has an added
            # replica dimension as the outermost dimension. On each replica it will
            # contain the local values and zeros for all other values that need to be
            # fetched from other replicas.
            ext_tensor = tf.scatter_nd(
                indices=[[replica_context.replica_id_in_sync_group]],
                updates=[tensor],
                shape=tf.concat([[num_replicas], tf.shape(tensor)], axis=0),
            )

            # As every value is only present on one replica and 0 in all others, adding
            # them all together will result in the full tensor on all replicas.
            ext_tensor = replica_context.all_reduce(tf.distribute.ReduceOp.SUM, ext_tensor)

            # Flatten the replica dimension.
            # The first dimension size will be: tensor.shape[0] * num_replicas
            # Using [-1] trick to support also scalar input.
            return tf.reshape(ext_tensor, [-1] + ext_tensor.shape.as_list()[2:])

    def get_config(self) -> dict[str, Any]:
        config = {
            "temperature": self.temperature,
        }
        base_config = super().get_config()
        return {**base_config, **config}



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Any

import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="Similarity")
class WarmupCosineDecay(tf.keras.optimizers.schedules.LearningRateSchedule):
    """A cosine decay LearningRateSchedule with a linear warmup period.

    See [Loshchilov & Hutter, ICLR2016](https://arxiv.org/abs/1608.03983),
    SGDR: Stochastic Gradient Descent with Warm Restarts.

    When training a model, it is often useful to lower the learning rate as
    the training progresses. This schedule applies a linear warmup from 0.0
    to a max learning rate followed by a cosine decay function. It requires
    a `step` value to compute the decayed learning rate. You can just pass
    a TensorFlow variable that you increment at each training step.

    The schedule is a 1-arg callable that produces a decayed learning rate
    when passed the current optimizer step. This can be useful for changing
    the learning rate value across different invocations of optimizer
    functions.

    It is computed as:

    ```python
    def decayed_learning_rate(step):
      step = min(step, total_steps)
      if step < warmup_steps:
        decayed = step / warmup_steps
      else:
        decay_steps = total_steps - warmup_steps
        cosine_decay = 0.5 * (1. + cos(pi * (step - warmup_steps) / decay_steps))
        decayed = (1 - alpha) * cosine_decay + alpha

      return max_learning_rate * decayed
    ```

    Example usage:
    ```python
    total_steps = 1000
    warmup_steps = 100
    lr_decayed_fn = tf.keras.optimizers.schedules.WarmupCosineDecay(
        max_learning_rate, total_steps, warmup_steps)
    ```

    You can pass this schedule directly into a `tf.keras.optimizers.Optimizer`
    as the learning rate.

    Returns:
      A 1-arg callable learning rate schedule that takes the current optimizer
      step and outputs the decayed learning rate, a scalar `Tensor` of the same
      type as `max_learning_rate`.
    """

    def __init__(
        self,
        max_learning_rate: float,
        total_steps: int,
        warmup_steps: int,
        alpha: float = 0.0,
        name: str | None = None,
    ):
        """Applies cosine decay with warmp up to the learning rate.

        Args:
          max_learning_rate: The max learning rate after warmup.
          total_steps: Total number of steps in the schedule.
          warmup_steps: Number of steps to warmup over. Must be smaller than the total number of steps.
          alpha: Minimum learning rate value as a fraction of initial_learning_rate.
          name: Optional name of the operation. Defaults to 'WarmupCosineDecay'.
        """
        super().__init__()

        if warmup_steps >= total_steps:
            raise ValueError("warmup_steps must be less than the total steps")

        self.max_learning_rate = max_learning_rate
        self.warmup_steps = warmup_steps
        self.total_steps = total_steps
        self.alpha = alpha
        self.name = name

        self.max_learning_rate_tf = tf.convert_to_tensor(self.max_learning_rate, name="max_learning_rate")
        self.dtype = self.max_learning_rate_tf.dtype
        self.warmup_steps_tf = tf.cast(self.warmup_steps, self.dtype)

        self.cosine_decay = tf.keras.experimental.CosineDecay(max_learning_rate, total_steps - warmup_steps, alpha)

    def __call__(self, step):
        with tf.name_scope(self.name or "WarmupCosineDecay"):
            step = tf.cast(step, self.dtype)

            learning_rate = tf.cond(
                tf.math.less(step, self.warmup_steps_tf),
                lambda: tf.math.divide_no_nan(step, self.warmup_steps_tf) * self.max_learning_rate_tf,
                lambda: self.cosine_decay(step - self.warmup_steps_tf),
            )

            return learning_rate

    def get_config(self) -> dict[str, Any]:
        return {
            "max_learning_rate": self.max_learning_rate,
            "total_steps": self.total_steps,
            "warmup_steps": self.warmup_steps,
            "alpha": self.alpha,
            "name": self.name,
        }



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import base64
import io
from collections.abc import Mapping, Sequence
from typing import Any

import numpy as np
import PIL
import umap
from bokeh import __version__
from bokeh.plotting import ColumnDataSource, figure, output_notebook, show
from distinctipy import distinctipy
from tqdm.auto import tqdm

from tensorflow_similarity.types import FloatTensor, Tensor


def tensor2images(tensor: Tensor, size: int = 64) -> list[str]:
    """Convert tensor images back to in memory images
    encoded in base 64.

    Args:
        tensor: 4D tensor that represent an image list.
        size: Image size to output in pixels. Defaults to 64.

    Returns:
        list of images encoded as base64 strings
    """

    # casting as iterating over a Tensor is slow.
    data = np.array(tensor)

    # if image provided are scaled between [0,1] then rescale
    if np.max(data) <= 1:
        data = data * 255

    # cast as int so PIL accepts its
    data = data.astype(np.uint8)

    imgs_b64 = []
    for a in tqdm(data, desc="generating diplayabe images"):
        # if single channel, treat it as black and white
        if a.shape[-1] == 1:
            a = np.reshape(a, (a.shape[0], a.shape[1]))
            img = PIL.Image.fromarray(a, "L")
        else:
            img = PIL.Image.fromarray(a)

        img_resized = img.resize((size, size))
        buffer = io.BytesIO()
        img_resized.save(buffer, format="JPEG")
        img_bytes = buffer.getvalue()
        img64 = "data:image/png;base64,%s" % str(base64.b64encode(img_bytes))[2:-1]
        imgs_b64.append(img64)

    return imgs_b64


def projector(
    embeddings: FloatTensor,
    labels: Sequence[Any] | None = None,
    class_mapping: Sequence[int] | None = None,
    images: Tensor | None = None,
    image_size: int = 64,
    tooltips_info: Mapping[str, Sequence[str]] | None = None,
    pt_size: int = 3,
    colorize: bool = True,
    pastel_factor: float = 0.1,
    plot_size: int = 600,
    active_drag: str = "box_zoom",
    densmap: bool = True,
):
    """Visualize the embeddings in 2D or 3D using UMAP projection

    Args:
        embeddings: The embeddings outputed by the model that
        are to be visualized

        labels: Labels associated with the embeddings. If not supplied treat
        each example as its own classes.

        class_mapping: Dictionary or list that maps the class numerical ids
        to their name.

        images: Images to display in tooltip on hover. Usually x_test tensor.

        image_size: size of the images displayed in the tool tip.
        Defaults to 64.

        pt_size: Size of the points displayed on the visualization.
        Defaults to 3.

        tooltips_info: Dictionary of information to display in the tooltips.

        colorize: Colorize the clusters. Defaults to true.

        pastel_factor: Modify the color palette to be more pastel.

        densmap: Use UMAP dense mapper which provides better density
        estimation but is a little slower. Defaults to True.
    """

    print("perfoming projection using UMAP")
    reducer = umap.UMAP(densmap=densmap)
    # FIXME: 2d vs 3d
    cords = reducer.fit_transform(embeddings)

    # sample id
    _idxs = list(range(len(embeddings)))

    # labels?
    if labels is not None:
        # if labels are already names just use them.
        if isinstance(labels[0], str):
            _labels = labels
        else:
            _labels = [int(i) for i in labels]
    else:
        # treat each examples as its own class
        _labels = _idxs

    # class name mapping?
    if class_mapping:
        _labels_txt = [class_mapping[i] for i in _labels]
    else:
        _labels_txt = [str(i) for i in _labels]

    class_list = sorted(set(_labels_txt))
    num_classes = len(class_list)

    # generate data
    data = dict(
        id=_idxs,
        x=[i[0] for i in cords],
        y=[i[1] for i in cords],
        labels=_labels,
        labels_txt=_labels_txt,
    )

    # colors if needed
    if labels is not None and colorize:
        # generate colors
        colors = {}
        for idx, c in enumerate(distinctipy.get_colors(num_classes, pastel_factor=pastel_factor)):
            # this is needed as labels can be strings or int or else
            cls_id = class_list[idx]
            colors[cls_id] = distinctipy.get_hex(c)

        # map point to their color
        _colors = [colors[i] for i in _labels_txt]
        data["colors"] = _colors
    else:
        _colors = []

    # building custom tooltips
    tooltips = '<div style="border:1px solid #ABABAB">'

    if images is not None:
        imgs = tensor2images(images, image_size)
        data["imgs"] = imgs
        # have to write custom tooltip html.
        tooltips += '<center><img src="@imgs"/></center>'  # noqa

    # adding user info
    if tooltips_info:
        for k, v in tooltips_info.items():
            data[k] = v
            tooltips += "%s:@%s <br>" % (k, k)

    tooltips += "Class:@labels_txt <br>ID:@id </div>"

    # to bokeh format
    source = ColumnDataSource(data=data)
    output_notebook()
    # Bokeh backward compatibility
    if int(__version__.split(".")[0]) >= 3:
        fig = figure(
            tooltips=tooltips,
            width=plot_size,
            height=plot_size,
            active_drag=active_drag,
            active_scroll="wheel_zoom",
        )
    else:
        fig = figure(
            tooltips=tooltips,
            plot_width=plot_size,
            plot_height=plot_size,
            active_drag=active_drag,
            active_scroll="wheel_zoom",
        )

    # remove grid and axis
    fig.xaxis.visible = False
    fig.yaxis.visible = False
    fig.xgrid.visible = False
    fig.ygrid.visible = False

    # draw points
    if len(_colors):
        fig.circle("x", "y", size=pt_size, color="colors", source=source)
    else:
        fig.circle("x", "y", size=pt_size, source=source)

    # render
    output_notebook()
    show(fig, notebook_handle=True)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import io
from collections.abc import Sequence
from pathlib import Path

import numpy as np
import pandas as pd
import tensorflow as tf

from tensorflow_similarity.types import FloatTensor, PandasDataFrame, Tensor

from .store import Store


class MemoryStore(Store):
    """Efficient in-memory dataset store"""

    def __init__(self) -> None:
        # We are using a native python array in memory for its row speed.
        # Serialization / export relies on Arrow.
        self.labels: list[int | None] = []
        self.embeddings: list[FloatTensor] = []
        self.data: list[Tensor | None] = []
        self.num_items: int = 0
        pass

    def add(
        self,
        embedding: FloatTensor,
        label: int | None = None,
        data: Tensor | None = None,
    ) -> int:
        """Add an Embedding record to the key value store.

        Args:
            embedding: Embedding predicted by the model.

            label: Class numerical id. Defaults to None.

            data: Data associated with the embedding. Defaults to None.

        Returns:
            Associated record id.
        """
        idx = self.num_items
        self.labels.append(label)
        self.embeddings.append(embedding)
        self.data.append(data)
        self.num_items += 1
        return idx

    def batch_add(
        self,
        embeddings: Sequence[FloatTensor],
        labels: Sequence[int] | None = None,
        data: Sequence[Tensor] | None = None,
    ) -> list[int]:
        """Add a set of embedding records to the key value store.

        Args:
            embeddings: Embeddings predicted by the model.

            labels: Class numerical ids. Defaults to None.

            data: Data associated with the embeddings. Defaults to None.

        See:
            add() for what a record contains.

        Returns:
            List of associated record id.
        """
        idxs: list[int] = []
        for idx, embedding in enumerate(embeddings):
            label = None if labels is None else labels[idx]
            rec_data = None if data is None else data[idx]
            idxs.append(self.add(embedding, label, rec_data))
        return idxs

    def get(self, idx: int) -> tuple[FloatTensor, int | None, Tensor | None]:
        """Get an embedding record from the key value store.

        Args:
            idx: Id of the record to fetch.

        Returns:
            record associated with the requested id.
        """

        return self.embeddings[idx], self.labels[idx], self.data[idx]

    def batch_get(self, idxs: Sequence[int]) -> tuple[list[FloatTensor], list[int | None], list[Tensor | None]]:
        """Get embedding records from the key value store.

        Args:
            idxs: ids of the records to fetch.

        Returns:
            List of records associated with the requested ids.
        """
        embeddings = []
        labels = []
        data = []
        for idx in idxs:
            e, l, d = self.get(idx)
            embeddings.append(e)
            labels.append(l)
            data.append(d)
        return embeddings, labels, data

    def size(self) -> int:
        "Number of record in the key value store."
        return self.num_items

    def save(self, path: str, compression: bool = True) -> None:
        """Serializes index on disk.

        Args:
            path: where to store the data.
            compression: Compress index data. Defaults to True.
        """
        # Writing to a buffer to avoid read error in np.savez when using GFile.
        # See: https://github.com/tensorflow/tensorflow/issues/32090
        io_buffer = io.BytesIO()
        if compression:
            np.savez_compressed(
                io_buffer,
                embeddings=self.embeddings,
                labels=np.array(self.labels),
                data=np.array(self.data),
            )
        else:
            np.savez(
                io_buffer,
                embeddings=self.embeddings,
                labels=np.array(self.labels),
                data=np.array(self.data),
            )

        with tf.io.gfile.GFile(self._make_fname(path), "wb+") as f:
            f.write(io_buffer.getvalue())

    def load(self, path: str) -> int:
        """load index on disk

        Args:
            path: which directory to use to store the index data.

        Returns:
           Number of records reloaded.
        """
        fname = self._make_fname(path, check_file_exit=True)
        with tf.io.gfile.GFile(fname, "rb") as gfp:
            data = np.load(gfp, allow_pickle=True)
        self.embeddings = list(data["embeddings"])
        self.labels = list(data["labels"])
        self.data = list(data["data"])
        self.num_items = len(self.embeddings)
        print("loaded %d records from %s" % (self.size(), path))
        return self.size()

    def _make_fname(self, path: str, check_file_exit: bool = False) -> str:
        p = Path(path)
        if not tf.io.gfile.exists(p):
            raise ValueError("Index path doesn't exist")
        fname = p / "index.npz"

        # only for loading
        if check_file_exit and not tf.io.gfile.exists(fname):
            raise ValueError("Index file not found")
        return str(fname)

    def to_data_frame(self, num_records: int = 0) -> PandasDataFrame:
        """Export data as a Pandas dataframe.

        Args:
            num_records: Number of records to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """

        if not num_records:
            num_records = self.num_items

        data = {
            "embeddings": self.embeddings[:num_records],
            "data": self.data[:num_records],
            "labels": self.labels[:num_records],
        }

        # forcing type from Any to PandasFrame
        df: PandasDataFrame = pd.DataFrame.from_dict(data)
        return df



import math
import sys
from unittest.mock import MagicMock

import numpy as np
import tensorflow as tf

from tensorflow_similarity import types, utils


def create_lookups():
    lookups = []
    for i in range(2):
        lookup_set = []
        for j in range(2):
            label = j + i * 2
            lookup_set.append(types.Lookup(rank=j, distance=float(j), label=label))

        lookups.append(lookup_set)

    return lookups


class UtilsTest(tf.test.TestCase):
    def setUp(self):
        super().setUp()
        # the lookups collection may be modified by the tests, so it must be
        # created in setUp before each test.
        self.lookups = create_lookups()

    def tearDown(self):
        super().tearDown()
        del self.lookups

    def test_tf_cap_memory_gpu_exists(self):
        tf.config.experimental.list_physical_devices = MagicMock(return_value=["foo"])
        tf.config.experimental.set_memory_growth = MagicMock()

        utils.tf_cap_memory()
        tf.config.experimental.set_memory_growth.assert_called_with("foo", True)

    def test_tf_cap_memory_no_gpu(self):
        tf.config.experimental.list_physical_devices = MagicMock(return_value=[])
        tf.config.experimental.set_memory_growth = MagicMock()

        utils.tf_cap_memory()
        tf.config.experimental.set_memory_growth.assert_not_called()

    def test_tf_cap_memory_runtime_error(self):
        tf.config.experimental.list_physical_devices = MagicMock(return_value=["foo"])
        tf.config.experimental.set_memory_growth = MagicMock(side_effect=RuntimeError("bar"))

        with self.captureWritesToStream(sys.stdout) as captured:
            utils.tf_cap_memory()

        self.assertRegex(captured.contents(), "bar")

    def test_is_tensor(self):
        self.assertTrue(utils.is_tensor_or_variable(tf.constant([0])))

    def test_is_variable(self):
        self.assertTrue(utils.is_tensor_or_variable(tf.Variable(1.0)))

    def test_is_not_tensor_or_variable(self):
        self.assertFalse(utils.is_tensor_or_variable([0]))

    def test_unpack_lookup_labels(self):
        with self.captureWritesToStream(sys.stdout) as captured:
            unpacked = utils.unpack_lookup_labels(self.lookups, dtype="int32")

        expected = tf.constant([[0, 1], [2, 3]], dtype="int32")

        self.assertAllEqual(unpacked, expected)
        self.assertEqual(captured.contents(), "")

    def test_unpack_lookup_labels_uneven_lookup_sets(self):
        # Add an extra label to the second lookup set
        self.lookups[1].append(types.Lookup(rank=3, distance=math.inf, label=4))

        with self.captureWritesToStream(sys.stdout) as captured:
            unpacked = utils.unpack_lookup_labels(self.lookups, dtype="int32")

        expected = tf.constant([[0, 1, 0x7FFFFFFF], [2, 3, 4]], dtype="int32")

        self.assertAllEqual(unpacked, expected)

        msg = (
            "WARNING: 1 lookup sets are shorter than the max lookup set "
            "length. Imputing 0x7FFFFFFF for the missing label lookups.\n"
        )

        self.assertEqual(captured.contents(), msg)

    def test_unpack_lookup_distances(self):
        with self.captureWritesToStream(sys.stdout) as captured:
            unpacked = utils.unpack_lookup_distances(self.lookups, dtype="float32")

        expected = tf.constant([[0.0, 1.0], [0.0, 1.0]], dtype="float32")

        self.assertAllEqual(unpacked, expected)
        self.assertEqual(captured.contents(), "")

    def test_unpack_lookup_distances_rounding(self):
        lookups = [[types.Lookup(rank=0, distance=0.11119, label=1)]]

        with self.captureWritesToStream(sys.stdout) as captured:
            unpacked = utils.unpack_lookup_distances(lookups, dtype="float32", distance_rounding=4)

        expected = tf.constant([[0.1112]], dtype="float32")

        self.assertAllEqual(unpacked, expected)
        self.assertEqual(captured.contents(), "")

    def test_unpack_lookup_distances_uneven_lookup_sets(self):
        # Add an extra label to the second lookup set
        self.lookups[1].append(types.Lookup(rank=3, distance=2.0, label=4))

        with self.captureWritesToStream(sys.stdout) as captured:
            unpacked = utils.unpack_lookup_distances(self.lookups, dtype="float32")

        expected = tf.constant([[0.0, 1.0, math.inf], [0.0, 1.0, 2.0]], dtype="float32")

        self.assertAllEqual(unpacked, expected)

        msg = (
            "WARNING: 1 lookup sets are shorter than the max lookup set "
            "length. Imputing math.inf for the missing distance lookups.\n"
        )

        self.assertEqual(captured.contents(), msg)

    def test_same_length_rows_check_same_length(self):
        x = tf.ragged.constant([[0, 1], [0, 2], [0, 3]])
        self.assertTrue(utils._same_length_rows(x))

    def test_same_length_rows_check_different_lengths(self):
        x = tf.ragged.constant([[0], [0, 2], [0, 2, 3]])
        self.assertFalse(utils._same_length_rows(x))

    def test_count_of_small_lookup_sets_ragged(self):
        x = tf.ragged.constant([[0], [0, 2], [0, 2, 3]])
        counts = utils._count_of_small_lookup_sets(x)
        self.assertEqual(counts, 2)

    def test_count_of_small_lookup_sets_all_same_length(self):
        x = tf.ragged.constant([[0, 2], [0, 2], [0, 2]])
        counts = utils._count_of_small_lookup_sets(x)
        self.assertEqual(counts, 0)

    def test_unpack_results(self):
        results = {"recall@1": np.array([1.0])}
        logs = {"some_other_callback": 0.0}
        tb_writer = MagicMock()
        tf.summary.scalar = MagicMock()

        mstr = utils.unpack_results(
            results=results,
            epoch=1,
            logs=logs,
            tb_writer=tb_writer,
            name_suffix="_test",
        )

        self.assertDictEqual(logs, {"some_other_callback": 0.0, "recall@1_test": 1.0})
        self.assertListEqual(mstr, ["recall@1_test: 1.0000"])
        tf.summary.scalar.assert_called_with("recall@1_test", 1.0, step=1)


if __name__ == "__main__":
    tf.test.main()



import tensorflow as tf


def generate_perfect_test_batch(batch_size=32):
    """Generate a batch of embeddings and labels that will result in a perfect loss score."""

    # y_true: labels
    y_true = tf.range(0, batch_size, dtype=tf.int32)
    y_true = tf.concat([y_true, y_true], axis=0)

    # y_preds: embedding
    y_preds = tf.one_hot(y_true, depth=batch_size, dtype=tf.float32)

    y_true = tf.expand_dims(y_true, axis=1)
    return y_true, y_preds


def generate_bad_test_batch(batch_size=32):
    """Generate a batch of embeddings and labels that will result in a mismatch for all classes."""

    # y_true: labels
    y_true = tf.range(0, batch_size, dtype=tf.int32)

    # y_preds: embedding
    y_preds = tf.concat([y_true, y_true[::-1]], axis=0)
    y_preds = tf.one_hot(y_preds, depth=batch_size, dtype=tf.float32)

    y_true = tf.concat([y_true, y_true], axis=0)
    y_true = tf.expand_dims(y_true, axis=1)
    return y_true, y_preds



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Mapping, MutableMapping, MutableSequence, Sequence
from copy import copy
from pathlib import Path
from typing import Any

import numpy as np
import tensorflow as tf
from tabulate import tabulate
from tensorflow.keras.losses import Loss
from tensorflow.keras.metrics import Metric
from tensorflow.keras.optimizers import Optimizer
from termcolor import cprint
from tqdm.auto import tqdm

from tensorflow_similarity.classification_metrics import (  # noqa
    ClassificationMetric,
    make_classification_metric,
)
from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.evaluators.evaluator import Evaluator
from tensorflow_similarity.indexer import Indexer
from tensorflow_similarity.layers import ActivationStdLoggingLayer
from tensorflow_similarity.losses import MetricLoss
from tensorflow_similarity.matchers import ClassificationMatch
from tensorflow_similarity.retrieval_metrics import RetrievalMetric
from tensorflow_similarity.search import Search
from tensorflow_similarity.stores import Store
from tensorflow_similarity.training_metrics import DistanceMetric
from tensorflow_similarity.types import (
    CalibrationResults,
    FloatTensor,
    IntTensor,
    Lookup,
    PandasDataFrame,
    Tensor,
)

# Value based on implementation from original papers.
BN_EPSILON = 1.001e-5


def get_projector(input_dim, dim=512, activation="relu", num_layers: int = 3):
    inputs = tf.keras.layers.Input((input_dim,), name="projector_input")
    x = inputs

    for i in range(num_layers - 1):
        x = tf.keras.layers.Dense(
            dim,
            use_bias=False,
            kernel_initializer=tf.keras.initializers.LecunUniform(),
            name=f"projector_layer_{i}",
        )(x)
        x = tf.keras.layers.BatchNormalization(epsilon=BN_EPSILON, name=f"batch_normalization_{i}")(x)
        x = tf.keras.layers.Activation(activation, name=f"{activation}_activation_{i}")(x)
    x = tf.keras.layers.Dense(
        dim,
        use_bias=False,
        kernel_initializer=tf.keras.initializers.LecunUniform(),
        name="projector_output",
    )(x)
    x = tf.keras.layers.BatchNormalization(
        epsilon=BN_EPSILON,
        center=False,  # Page:5, Paragraph:2 of SimSiam paper
        scale=False,  # Page:5, Paragraph:2 of SimSiam paper
        name="batch_normalization_ouput",
    )(x)
    # Metric Logging layer. Monitors the std of the layer activations.
    # Degnerate solutions colapse to 0 while valid solutions will move
    # towards something like 0.0220. The actual number will depend on the layer size.
    outputs = ActivationStdLoggingLayer(name="proj_std")(x)
    projector = tf.keras.Model(inputs, outputs, name="projector")
    return projector


def get_predictor(input_dim, hidden_dim=512, activation="relu"):
    inputs = tf.keras.layers.Input(shape=(input_dim,), name="predictor_input")
    x = inputs

    x = tf.keras.layers.Dense(
        hidden_dim,
        use_bias=False,
        kernel_initializer=tf.keras.initializers.LecunUniform(),
        name="predictor_layer_0",
    )(x)
    x = tf.keras.layers.BatchNormalization(epsilon=BN_EPSILON, name="batch_normalization_0")(x)
    x = tf.keras.layers.Activation(activation, name=f"{activation}_activation_0")(x)

    x = tf.keras.layers.Dense(
        input_dim,
        kernel_initializer=tf.keras.initializers.LecunUniform(),
        name="predictor_output",
    )(x)
    # Metric Logging layer. Monitors the std of the layer activations.
    # Degnerate solutions colapse to 0 while valid solutions will move
    # towards something like 0.0220. The actual number will depend on the layer size.
    outputs = ActivationStdLoggingLayer(name="pred_std")(x)
    predictor = tf.keras.Model(inputs, outputs, name="predictor")
    return predictor


def create_contrastive_model(
    *args,
    backbone: tf.keras.Model,
    projector: tf.keras.Model | None = None,
    predictor: tf.keras.Model | None = None,
    algorithm: str = "simsiam",
    **kwargs,
) -> ContrastiveModel:
    """Create a contrastive model."""
    if projector is None:
        projector = get_projector(input_dim=backbone.output_shape[-1], num_layers=2)

    input_shape = backbone.input_shape[1:]
    inputs = tf.keras.layers.Input(shape=input_shape, name="main_model_input")
    projector_features = projector(backbone(inputs))
    if algorithm == "simsiam":
        if predictor is None:
            predictor = get_predictor(input_dim=projector.output_shape[-1])
        predictor_features = predictor(projector_features)
    else:
        predictor_features = None

    outputs = [projector_features]
    if predictor_features is not None:
        outputs.append(predictor_features)

    return ContrastiveModel(
        *args,
        backbone=backbone,
        projector=projector,
        predictor=predictor,
        algorithm=algorithm,
        inputs=inputs,
        outputs=outputs,
        **kwargs,
    )


@tf.keras.utils.register_keras_serializable(package="Similarity")
class ContrastiveModel(tf.keras.Model):
    def __init__(
        self,
        *args,
        backbone: tf.keras.Model,
        projector: tf.keras.Model,
        predictor: tf.keras.Model | None = None,
        algorithm: str = "simsiam",
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.backbone = backbone
        self.projector = projector
        self.predictor = predictor

        self.algorithm = algorithm

        self._create_loss_trackers()

        self.supported_algorithms = ("simsiam", "simclr", "barlow", "vicreg")

        if self.algorithm not in self.supported_algorithms:
            raise ValueError(
                f"{self.algorithm} is not a supported algorithm."
                f"Supported algorithms are {self.supported_algorithms}."
            )

    def compile(
        self,
        optimizer: Optimizer | str | Mapping | Sequence = "rmsprop",
        loss: Loss | MetricLoss | str | Mapping | Sequence | None = None,
        metrics: Metric | DistanceMetric | str | Mapping | Sequence | None = None,  # noqa
        loss_weights: Mapping | Sequence | None = None,
        weighted_metrics: Metric | DistanceMetric | str | Mapping | Sequence | None = None,  # noqa
        run_eagerly: bool = False,
        steps_per_execution: int = 1,
        distance: Distance | str = "cosine",
        kv_store: Store | str = "memory",
        search: Search | str = "nmslib",
        evaluator: Evaluator | str = "memory",
        stat_buffer_size: int = 1000,
        **kwargs,
    ):
        """Configures the model for training.

        Args:
            optimizer: String (name of optimizer) or optimizer instance. See
              [tf.keras.optimizers](https://www.tensorflow.org/api_docs/python/tf/keras/optimizers).

            loss: String (name of objective function), objective function, any
              `tensorflow_similarity.loss.*` instance or a `tf.keras.losses.Loss`
              instance. See the [Losses documentation](../losses.md) for a list of
              metric learning specific losses offered by TensorFlow Similarity and
              [tf.keras.losses](https://www.tensorflow.org/api_docs/python/tf/keras/losses)
              for the losses available directly in TensorFlow.

            metrics: List of metrics to be evaluated by the model during
              training and testing. Each of those can be a string, a function or a
              [tensorflow_similarity.metrics.*](../metrics.md) instance. Note that
              the metrics used for some types of metric-learning such as distance
              learning (e.g via triplet loss) have a different prototype than the
              metrics used in standard models and you can't use the
              `tf.keras.metrics` for those types of learning.

              Additionally many distance metrics are computed based of the
              [Indexer()](../indexer.md) performance. E.g Matching Top 1 accuracy.
              For technical and performance reasons, indexing data at each
              training batch to compute those is impractical so those metrics are
              computed at epoch end via the [EvalCallback](../callbacks.md)

              See [Evaluation Metrics](../eval_metrics.md) for a list of available
              metrics.

            loss_weights: Optional list or dictionary specifying scalar
              coefficients (Python floats) to weight the loss contributions of
              different model outputs. The loss value that will be minimized by
              the model will then be the *weighted sum* of all individual losses,
              weighted by the `loss_weights` coefficients.  If a list, it is
              expected to have a 1:1 mapping to the model's outputs. If a dict, it
              is expected to map output names (strings) to scalar coefficients.

            weighted_metrics: List of metrics to be evaluated and weighted by
              sample_weight or class_weight during training and testing.

            run_eagerly: Bool. Defaults to `False`. If `True`, this `Model`'s
              logic will not be wrapped in a `tf.function`. Recommended to leave
              this as `None` unless your `Model` cannot be run inside a
              `tf.function`.

            steps_per_execution: Int. Defaults to 1. The number of batches to
              run during each `tf.function` call. Running multiple batches inside
              a single `tf.function` call can greatly improve performance on TPUs
              or small models with a large Python overhead.  At most, one full
              epoch will be run each execution. If a number larger than the size
              of the epoch is passed,  the execution will be truncated to the size
              of the epoch.  Note that if `steps_per_execution` is set to `N`,
              `Callback.on_batch_begin` and `Callback.on_batch_end` methods will
              only be called every `N` batches (i.e. before/after each
              `tf.function` execution).

            distance: Distance used to compute embeddings proximity.  Defaults
              to 'cosine'.

            kv_store: How to store the indexed records.  Defaults to 'memory'.

            search: Which `Search()` framework to use to perform KNN search.
              Defaults to 'nmslib'.

            evaluator: What type of `Evaluator()` to use to evaluate index
              performance. Defaults to in-memory one.

            stat_buffer_size: Size of the sliding windows buffer used to compute
              index performance. Defaults to 1000.

        Raises:
            ValueError: In case of invalid arguments for
                `optimizer`, `loss` or `metrics`.
        """
        distance_obj = distance_canonicalizer(distance)

        # init index
        self.create_index(
            distance=distance_obj,
            search=search,
            kv_store=kv_store,
            evaluator=evaluator,
            stat_buffer_size=stat_buffer_size,
        )

        # call underlying keras method
        super().compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            loss_weights=loss_weights,
            weighted_metrics=weighted_metrics,
            run_eagerly=run_eagerly,
            steps_per_execution=steps_per_execution,
            **kwargs,
        )

    def _create_loss_trackers(self):
        self.loss_trackers = {}
        self.loss_trackers["loss"] = tf.keras.metrics.Mean(name="loss")

        self.reg_loss_len = len(self.backbone.losses) + len(self.projector.losses)
        if self.predictor is not None:
            self.reg_loss_len += len(self.predictor.losses)

        if self.reg_loss_len:
            self.loss_trackers["contrastive_loss"] = tf.keras.metrics.Mean(name="contrastive_loss")
            self.loss_trackers["regularization_loss"] = tf.keras.metrics.Mean(name="regularization_loss")

    @property
    def metrics(self):
        # We remove the compiled loss metric because we want to manually track
        # the loss in train_step and test_step.
        base_metrics = [m for m in super().metrics if m.name != "loss"]
        loss_trackers = list(self.loss_trackers.values())
        loss_trackers.extend(base_metrics)
        return loss_trackers

    def train_step(self, data):
        view1, view2 = self._parse_views(data)

        # Forward pass through the backbone, projector, and predictor
        with tf.GradientTape() as tape:
            contrastive_loss, pred1, pred2, z1, z2 = self._forward_pass(view1, view2, training=True)
            regularization_loss = sum(self.backbone.losses)
            regularization_loss += sum(self.projector.losses)
            if self.predictor is not None:
                regularization_loss += sum(self.predictor.losses)
            combined_loss = contrastive_loss + regularization_loss

        # collect train variables from both the backbone, projector, and projector
        tvars = self.backbone.trainable_variables
        tvars = tvars + self.projector.trainable_variables

        if self.predictor is not None:
            tvars = tvars + self.predictor.trainable_variables

        # Compute gradients
        gradients = tape.gradient(combined_loss, tvars)

        # Update weights
        self.optimizer.apply_gradients(zip(gradients, tvars))

        # Update metrics
        # !This are contrastive metrics with different input
        # TODO: figure out interesting metrics -- z Mae?
        # TODO: check metrics are of the right type in compile?
        # self.compiled_metrics.update_state([z1, z2], [pred2, pred1])

        # report loss manually
        self.loss_trackers["loss"].update_state(combined_loss)
        if self.reg_loss_len:
            self.loss_trackers["contrastive_loss"].update_state(contrastive_loss)
            self.loss_trackers["regularization_loss"].update_state(regularization_loss)

        # Collect metrics to return
        return_metrics = {}
        for metric in self.metrics:
            result = metric.result()
            if isinstance(result, dict):
                return_metrics.update(result)
            else:
                return_metrics[metric.name] = result
        return return_metrics

    def test_step(self, data):
        view1, view2 = self._parse_views(data)

        contrastive_loss, pred1, pred2, z1, z2 = self._forward_pass(view1, view2, training=False)
        regularization_loss = sum(self.backbone.losses)
        regularization_loss += sum(self.projector.losses)
        if self.predictor is not None:
            regularization_loss += sum(self.predictor.losses)
        combined_loss = contrastive_loss + regularization_loss

        # Update metrics
        # !This are contrastive metrics with different input
        # TODO: figure out interesting metrics -- z Mae?
        # TODO: check metrics are of the right type in compile?
        # self.compiled_metrics.update_state([z1, z2], [pred2, pred1])

        # report loss manually
        self.loss_trackers["loss"].update_state(combined_loss)
        if self.reg_loss_len:
            self.loss_trackers["contrastive_loss"].update_state(contrastive_loss)
            self.loss_trackers["regularization_loss"].update_state(regularization_loss)

        # Collect metrics to return
        return_metrics = {}
        for metric in self.metrics:
            result = metric.result()
            if isinstance(result, dict):
                return_metrics.update(result)
            else:
                return_metrics[metric.name] = result
        return return_metrics

    def _forward_pass(self, view1, view2, training):
        h1 = self.backbone(view1, training=training)
        h2 = self.backbone(view2, training=training)

        z1 = self.projector(h1, training=training)
        z2 = self.projector(h2, training=training)

        if self.algorithm == "simsiam":
            p1 = self.predictor(z1, training=training)
            p2 = self.predictor(z2, training=training)
            l1 = self.compiled_loss(tf.stop_gradient(z1), p2)
            l2 = self.compiled_loss(tf.stop_gradient(z2), p1)
            loss = l1 + l2
            pred1, pred2 = p1, p2
        elif self.algorithm in ("simclr", "barlow", "vicreg"):
            loss = self.compiled_loss(z1, z2)
            pred1, pred2 = z1, z2

        return loss, pred1, pred2, z1, z2

    def _parse_views(self, data: Sequence[FloatTensor]) -> tuple[FloatTensor, FloatTensor]:
        if len(data) == 2:
            view1, view2 = data
        else:
            view1 = view2 = data[0]

        return view1, view2

    # fix TF 2.x < 2.7 bugs when using generator
    # todo(ovallis): link to original issue.
    def call(self, inputs, training=None, mask=None):
        return inputs

    def summary(self):
        cprint("[Backbone]", "green")
        self.backbone.summary()
        cprint("\n[Projector]", "magenta")
        self.projector.summary()
        if self.predictor is not None:
            cprint("\n[Predictor]", "magenta")
            self.projector.summary()

    def predict(
        self,
        x: FloatTensor,
        batch_size: int | None = None,
        verbose: int = 0,
        steps: int | None = None,
        callbacks: tf.keras.callbacks.Callback | None = None,
        max_queue_size: int = 10,
        workers: int = 1,
        use_multiprocessing: bool = False,
    ) -> FloatTensor:
        """Generates output predictions for the input samples.

        The predict returns the L2 normalized output of the projector.

        This can be used for indexing and querying examples, and is used by the
        EvalCallback.
        """
        x = self.backbone.predict(
            x,
            batch_size,
            verbose,
            steps,
            callbacks,
            max_queue_size,
            workers,
            use_multiprocessing,
        )
        x = self.projector.predict(
            x,
            batch_size,
            verbose,
            steps,
            callbacks,
            max_queue_size,
            workers,
            use_multiprocessing,
        )

        output: FloatTensor = tf.math.l2_normalize(x, axis=1)

        return output

    def create_index(
        self,
        distance: Distance | str = "cosine",
        search: Search | str = "nmslib",
        kv_store: Store | str = "memory",
        evaluator: Evaluator | str = "memory",
        stat_buffer_size: int = 1000,
    ) -> None:
        """Create the model index to make embeddings searchable via KNN.

        This method is normally called as part of `SimilarityModel.compile()`.
        However, this method is provided if users want to define a custom index
        outside of the `compile()` method.

        NOTE: This method sets `SimilarityModel._index` and will replace any
        existing index.

        Args:
            distance: Distance used to compute embeddings proximity. Defaults to
            'auto'.

            kv_store: How to store the indexed records.  Defaults to 'memory'.

            search: Which `Search()` framework to use to perform KNN search.
            Defaults to 'nmslib'.

            evaluator: What type of `Evaluator()` to use to evaluate index
            performance. Defaults to in-memory one.

            stat_buffer_size: Size of the sliding windows buffer used to compute
            index performance. Defaults to 1000.

        Raises:
            ValueError: Invalid search framework or key value store.
        """
        self._index = Indexer(
            embedding_size=self.projector.output_shape[-1],
            distance=distance,
            search=search,
            kv_store=kv_store,
            evaluator=evaluator,
            embedding_output=None,
            stat_buffer_size=stat_buffer_size,
        )

    def index(
        self,
        x: Tensor,
        y: IntTensor | None = None,
        data: Tensor | None = None,
        build: bool = True,
        verbose: int = 1,
    ):
        """Index data.

        Args:
            x: Samples to index.

            y: class ids associated with the data if any. Defaults to None.

            store_data: store the data associated with the samples in the key
            value store. Defaults to True.

            build: Rebuild the index after indexing. This is needed to make the
            new samples searchable. Set it to false to save processing time
            when calling indexing repeatidly without the need to search between
            the indexing requests. Defaults to True.

            verbose: Output indexing progress info. Defaults to 1.
        """

        if not self._index:
            raise Exception("You need to compile the model with a valid" "distance to be able to use the indexing")
        if verbose:
            print("[Indexing %d points]" % len(x))
            print("|-Computing embeddings")
        predictions = self.predict(x)

        self._index.batch_add(
            predictions=predictions,
            labels=y,
            data=data,
            build=build,
            verbose=verbose,
        )

    def index_single(
        self,
        x: Tensor,
        y: IntTensor | None = None,
        data: Tensor | None = None,
        build: bool = True,
        verbose: int = 1,
    ):
        """Index data.

        Args:
            x: Sample to index.

            y: class id associated with the data if any. Defaults to None.

            data: store the data associated with the samples in the key
            value store. Defaults to None.

            build: Rebuild the index after indexing. This is needed to make the
            new samples searchable. Set it to false to save processing time
            when calling indexing repeatidly without the need to search between
            the indexing requests. Defaults to True.

            verbose: Output indexing progress info. Defaults to 1.
        """

        if not self._index:
            raise Exception("You need to compile the model with a valid" "distance to be able to use the indexing")
        if verbose:
            print("[Indexing 1 point]")
            print("|-Computing embeddings")

        x = tf.expand_dims(x, axis=0)
        prediction = self.predict(x)
        self._index.add(
            prediction=prediction,
            label=y,
            data=data,
            build=build,
            verbose=verbose,
        )

    def lookup(self, x: Tensor, k: int = 5, verbose: int = 1) -> list[list[Lookup]]:
        """Find the k closest matches in the index for a set of samples.

        Args:
            x: Samples to match.

            k: Number of nearest neighboors to lookup. Defaults to 5.

            verbose: display progress. Default to 1.

        Returns
            list of list of k nearest neighboors:
            list[list[Lookup]]
        """
        predictions = self.predict(x)
        return self._index.batch_lookup(predictions=predictions, k=k, verbose=verbose)

    def single_lookup(self, x: Tensor, k: int = 5) -> list[Lookup]:
        """Find the k closest matches in the index for a given sample.

        Args:
            x: Sample to match.

            k: Number of nearest neighboors to lookup. Defaults to 5.

        Returns
            list of the k nearest neigboors info:
            list[Lookup]
        """
        x = tf.expand_dims(x, axis=0)
        prediction = self.predict(x)
        return self._index.single_lookup(prediction=prediction, k=k)

    def index_summary(self):
        "Display index info summary."
        self._index.print_stats()

    def calibrate(
        self,
        x: FloatTensor,
        y: IntTensor,
        thresholds_targets: MutableMapping[str, float] = {},
        k: int = 1,
        calibration_metric: str | ClassificationMetric = "f1",
        matcher: str | ClassificationMatch = "match_nearest",
        extra_metrics: MutableSequence[str | ClassificationMetric] = [
            "precision",
            "recall",
        ],  # noqa
        rounding: int = 2,
        verbose: int = 1,
    ) -> CalibrationResults:
        """Calibrate model thresholds using a test dataset.

        TODO: more detailed explaination.

        Args:

            x: examples to use for the calibration.

            y: labels associated with the calibration examples.

            thresholds_targets: Dict of performance targets to (if possible)
            meet with respect to the `calibration_metric`.

            calibration_metric:
            [ClassificationMetric()](classification_metrics/overview.md) used
            to evaluate the performance of the index.

            k: How many neighboors to use during the calibration.
            Defaults to 1.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold. Defaults to
            'match_nearest'.

            extra_metrics: List of additional
            `tf.similarity.classification_metrics.ClassificationMetric()`
            to compute and report. Defaults to ['precision', 'recall'].


            rounding: Metric rounding. Default to 2 digits.

            verbose: Be verbose and display calibration results.
            Defaults to 1.

        Returns:
            CalibrationResults containing the thresholds and cutpoints Dicts.
        """

        # predict
        predictions = self.predict(x)

        # calibrate
        return self._index.calibrate(
            predictions=predictions,
            target_labels=y,
            thresholds_targets=thresholds_targets,
            k=k,
            calibration_metric=calibration_metric,
            matcher=matcher,
            extra_metrics=extra_metrics,
            rounding=rounding,
            verbose=verbose,
        )

    def match(
        self,
        x: FloatTensor,
        cutpoint="optimal",
        no_match_label=-1,
        k=1,
        matcher: str | ClassificationMatch = "match_nearest",
        verbose=0,
    ):
        """Match a set of examples against the calibrated index

        For the match function to work, the index must be calibrated using
        calibrate().

        Args:
            x: Batch of examples to be matched against the index.

            cutpoint: Which calibration threshold to use.
            Defaults to 'optimal' which is the optimal F1 threshold computed
            using calibrate().

            no_match_label: Which label value to assign when there is no
            match. Defaults to -1.

            k: How many neighboors to use during the calibration.
            Defaults to 1.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            verbose. Be verbose. Defaults to 0.

        Returns:
            List of class ids that matches for each supplied example

        Notes:
            This function matches all the cutpoints at once internally as there
            is little performance downside to do so and allows to do the
            evaluation in a single go.

        """
        # basic checks
        if not self._index.is_calibrated:
            raise ValueError("Uncalibrated model: run model.calibration()")

        # get predictions
        predictions = self.predict(x)

        # matching
        matches = self._index.match(
            predictions,
            no_match_label=no_match_label,
            k=k,
            matcher=matcher,
            verbose=verbose,
        )

        # select which matches to return
        if cutpoint == "all":  # returns all the cutpoints for eval purpose.
            return matches
        else:  # normal match behavior - returns a specific cut point
            return matches[cutpoint]

    def evaluate_retrieval(
        self,
        x: Tensor,
        y: IntTensor,
        retrieval_metrics: Sequence[RetrievalMetric],  # noqa
        verbose: int = 1,
    ) -> dict[str, np.ndarray]:
        """Evaluate the quality of the index against a test dataset.

        Args:
            x: Examples to be matched against the index.

            y: Label associated with the examples supplied.

            retrieval_metrics: List of
            [RetrievalMetric()](retrieval_metrics/overview.md) to compute.

            verbose (int, optional): Display results if set to 1 otherwise
            results are returned silently. Defaults to 1.

        Returns:
            Dictionary of metric results where keys are the metric names and
            values are the metrics values.

        Raises:
            IndexError: Index must contain embeddings but is currently empty.
        """
        if self._index.size() == 0:
            raise IndexError("Index must contain embeddings but is " "currently empty. Have you run model.index()?")

        # get embeddings
        if verbose:
            print("|-Computing embeddings")
        predictions = self.predict(x)

        if verbose:
            print("|-Computing retrieval metrics")

        results = self._index.evaluate_retrieval(
            predictions=predictions,
            target_labels=y,
            retrieval_metrics=retrieval_metrics,
            verbose=verbose,
        )

        if verbose:
            table = zip(results.keys(), results.values())
            headers = ["metric", "Value"]
            print("\n [Summary]\n")
            print(tabulate(table, headers=headers))

        return results

    def evaluate_classification(
        self,
        x: Tensor,
        y: IntTensor,
        k: int = 1,
        extra_metrics: MutableSequence[str | ClassificationMetric] = [
            "precision",
            "recall",
        ],  # noqa
        matcher: str | ClassificationMatch = "match_nearest",
        verbose: int = 1,
    ) -> defaultdict[str, dict[str, str | np.ndarray]]:
        """Evaluate model classification matching on a given evaluation dataset.

        Args:
            x: Examples to be matched against the index.

            y: Label associated with the examples supplied.

            k: How many neighbors to use to perform the evaluation.
            Defaults to 1.

            extra_metrics: List of additional
            `tf.similarity.classification_metrics.ClassificationMetric()` to
            compute and report. Defaults to ['precision', 'recall'].

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            verbose (int, optional): Display results if set to 1 otherwise
            results are returned silently. Defaults to 1.

        Returns:
            Dictionary of (distance_metrics.md)[evaluation metrics]

        Raises:
            IndexError: Index must contain embeddings but is currently empty.
            ValueError: Uncalibrated model: run model.calibration()")
        """
        # There is some code duplication in this function but that is the best
        # solution to keep the end-user API clean and doing inferences once.
        if self._index.size() == 0:
            raise IndexError("Index must contain embeddings but is " "currently empty. Have you run model.index()?")

        if not self._index.is_calibrated:
            raise ValueError("Uncalibrated model: run model.calibration()")

        cal_metric = self._index.get_calibration_metric()

        # get embeddings
        if verbose:
            print("|-Computing embeddings")
        predictions = self.predict(x)

        results: defaultdict[str, dict[str, str | np.ndarray]] = defaultdict(dict)

        if verbose:
            pb = tqdm(total=len(self._index.cutpoints), desc="Evaluating cutpoints")

        for cp_name, cp_data in self._index.cutpoints.items():
            # create a metric that match at the requested k and threshold
            distance_threshold = float(cp_data["distance"])
            metric = make_classification_metric(cal_metric.name)
            metrics = copy(extra_metrics)
            metrics.append(metric)

            res: dict[str, str | np.ndarray] = {}
            res.update(
                self._index.evaluate_classification(
                    predictions,
                    y,
                    [distance_threshold],
                    metrics=metrics,
                    matcher=matcher,
                    k=k,
                )
            )
            res["distance"] = tf.constant([distance_threshold])
            res["name"] = cp_name
            results[cp_name] = res
            if verbose:
                pb.update()

        if verbose:
            pb.close()

        if verbose:
            headers = ["name", cal_metric.name]
            for i in results["optimal"].keys():
                if i not in headers:
                    headers.append(str(i))
            rows = []
            for data in results.values():
                rows.append([data[v] for v in headers])
            print("\n [Summary]\n")
            print(tabulate(rows, headers=headers))

        return results

    def reset_index(self):
        "Reinitialize the index"
        self._index.reset()

    def index_size(self) -> int:
        "Return the index size"
        return self._index.size()

    def load_index(self, filepath: str):
        """Load Index data from a checkpoint and initialize underlying
        structure with the reloaded data.

        Args:
            path: Directory where the checkpoint is located.
            verbose: Be verbose. Defaults to 1.
        """

        index_path = Path(filepath) / "index"
        self._index = Indexer.load(index_path)

    def save_index(self, filepath, compression=True):
        """Save the index to disk

        Args:
            path: directory where to save the index
            compression: Store index data compressed. Defaults to True.
        """
        index_path = Path(filepath) / "index"
        self._index.save(index_path, compression=compression)

    def save(
        self,
        filepath: str | Path,
        save_index: bool = True,
        compression: bool = True,
        overwrite: bool = True,
        include_optimizer: bool = True,
        save_format: str | None = None,
        signatures: Callable | Mapping[str, Callable] | None = None,
        options: tf.saved_model.SaveOptions | None = None,
        save_traces: bool = True,
    ) -> None:
        """Save Constrative model backbone, projector, and predictor

        Args:
            filepath: where to save the model.
            save_index: Save the index content. Defaults to True.
            compression: Compress index data. Defaults to True.
            overwrite: Overwrite previous model. Defaults to True.
            include_optimizer: Save optimizer state. Defaults to True.
            save_format: Either 'tf' or 'h5', indicating whether to save the
              model to Tensorflow SavedModel or HDF5. Defaults to 'tf' in
              TF 2.X, and 'h5' in TF 1.X.
            signatures: Signatures to save with the SavedModel. Applicable to
              the 'tf' format only. Please see the signatures argument in
              tf.saved_model.save for details.
            options: A `tf.saved_model.SaveOptions` to save with the model.
              Defaults to None.
            save_traces (optional): When enabled, the SavedModel will store the
              function traces for each layer. This can be disabled, so that only
              the configs of each layer are stored.  Defaults to True. Disabling
              this will decrease serialization time and reduce file size, but it
              requires that all custom layers/models implement a get_config()
              method.
        """
        super().save(
            filepath,
            overwrite=overwrite,
            include_optimizer=include_optimizer,
            save_format=save_format,
            signatures=signatures,
            options=options,
            save_traces=save_traces,
        )

        if hasattr(self, "_index") and self._index and save_index:
            self.save_index(filepath, compression=compression)
        else:
            msg = "The index was not saved with the model."
            if not hasattr(self, "_index"):
                msg = msg + (
                    "The model does not currently have an index. To use indexing "
                    "you must call either model.compile() or model.create_index() "
                    "and set a valid Distance."
                )

            if not save_index:
                msg = msg + " The save_index param is set to False."

            print(msg)

    def to_data_frame(self, num_items: int = 0) -> PandasDataFrame:
        """Export data as pandas dataframe

        Args:
            num_items (int, optional): Num items to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """
        return self._index.to_data_frame(num_items=num_items)

    def get_config(self) -> dict[str, Any]:
        config = {
            "backbone": self.backbone,
            "projector": self.projector,
            "predictor": self.predictor,
            "algorithm": self.algorithm,
        }
        base_config = super().get_config()
        return {**base_config, **config}

    @classmethod
    def from_config(cls, config):
        if "layers" in config:
            del config["layers"]
        if "input_layers" in config:
            del config["input_layers"]
        if "output_layers" in config:
            del config["output_layers"]
        return create_contrastive_model(**config)



import tensorflow as tf
from tensorflow.keras.losses import Reduction
from tensorflow.python.framework import combinations

from tensorflow_similarity import losses


# TODO(ovallis): Refactor XBM loss to work with graph mode and update tests.
@combinations.generate(
    combinations.combine(
        mode=["eager"],
    )
)
class XBMLossTest(tf.test.TestCase):
    def test_config(self):
        xbm_tpl_obj = losses.XBM(
            loss=losses.TripletLoss(distance="cosine"),
            memory_size=12,
            reduction=Reduction.SUM,
            name="xbm_triplet_loss",
        )
        self.assertEqual(xbm_tpl_obj.distance.name, "cosine")
        self.assertEqual(xbm_tpl_obj.name, "xbm_triplet_loss")
        self.assertEqual(xbm_tpl_obj.memory_size, 12)
        self.assertEqual(xbm_tpl_obj.reduction, Reduction.SUM)

    def test_xbm_loss(self):
        batch_size = 6
        embed_dim = 16

        embeddings1 = tf.random.uniform(shape=[batch_size, embed_dim])
        labels1 = tf.constant(
            [[1], [1], [2], [2], [3], [3]],
            dtype=tf.int32,
        )

        embeddings2 = tf.random.uniform(shape=[batch_size, embed_dim])
        labels2 = tf.constant(
            [[4], [4], [5], [5], [6], [6]],
            dtype=tf.int32,
        )

        distance = "cosine"
        loss = losses.MultiSimilarityLoss(distance=distance)
        loss_nowarm = losses.XBM(loss, memory_size=12, warmup_steps=0)

        # test enqueue
        loss_nowarm(labels1, embeddings1)
        self.assertAllEqual(loss_nowarm._y_pred_memory.numpy().shape, (batch_size, embed_dim))
        self.assertAllEqual(loss_nowarm._y_true_memory, labels1)

        loss_nowarm(labels2, embeddings2)
        self.assertAllEqual(loss_nowarm._y_pred_memory.numpy().shape, (2 * batch_size, embed_dim))
        self.assertAllEqual(loss_nowarm._y_true_memory, tf.concat([labels2, labels1], axis=0))

        # test dequeue
        loss_nowarm(labels2, embeddings2)
        self.assertAllEqual(loss_nowarm._y_pred_memory.numpy().shape, (2 * batch_size, embed_dim))
        self.assertAllEqual(loss_nowarm._y_true_memory, tf.concat([labels2, labels2], axis=0))

        # test warmup
        loss_warm = losses.XBM(loss, memory_size=12, warmup_steps=1)

        loss_warm(labels1, embeddings1)
        self.assertAllEqual(loss_warm._y_pred_memory.numpy().shape, (0, embed_dim))
        self.assertAllEqual(loss_warm._y_true_memory, tf.constant([[]], shape=(0, 1), dtype=tf.int32))

        loss_warm(labels2, embeddings2)
        self.assertAllEqual(loss_warm._y_pred_memory.numpy().shape, (batch_size, embed_dim))
        self.assertAllEqual(loss_warm._y_true_memory, labels2)



import tensorflow as tf

from tensorflow_similarity import types


def basic_lookup(rank=1, distance=0.1):
    return types.Lookup(rank=rank, distance=distance)


def label_lookup(label=10):
    return types.Lookup(rank=1, distance=0.1, label=label)


def embedding_lookup(embedding):
    return types.Lookup(rank=1, distance=0.1, embedding=embedding)


def data_lookup(data):
    return types.Lookup(rank=1, distance=0.1, data=data)


class TypesTest(tf.test.TestCase):
    def test_basic_lookup_eq(self):
        l1 = basic_lookup()
        l2 = basic_lookup()

        self.assertEqual(l1, l2)

    def test_basic_lookup_rank_not_eq(self):
        l1 = basic_lookup()
        l2 = basic_lookup(rank=2)

        self.assertNotEqual(l1, l2)

    def test_basic_lookup_distance_not_eq(self):
        l1 = basic_lookup()
        l2 = basic_lookup(distance=0.3)

        self.assertNotEqual(l1, l2)

    def test_label_lookup_eq(self):
        l1 = label_lookup()
        l2 = label_lookup()

        self.assertEqual(l1, l2)

    def test_label_lookup_not_eq(self):
        l1 = label_lookup()
        l2 = label_lookup(label=20)

        self.assertNotEqual(l1, l2)

    def test_optional_label_lookup_not_eq(self):
        l1 = label_lookup()
        l2 = label_lookup(label=None)

        self.assertNotEqual(l1, l2)

    def test_embedding_lookup_eq(self):
        l1 = embedding_lookup(embedding=tf.ones((1, 512)))
        l2 = embedding_lookup(embedding=tf.ones((1, 512)))

        self.assertEqual(l1, l2)

    def test_embedding_lookup_not_eq(self):
        l1 = embedding_lookup(embedding=tf.ones((1, 512)))
        l2 = embedding_lookup(embedding=tf.ones((1, 512)) + 1)

        self.assertNotEqual(l1, l2)

    def test_optional_embedding_lookup_not_eq(self):
        l1 = embedding_lookup(embedding=tf.ones((1, 512)))
        l2 = embedding_lookup(embedding=None)

        self.assertNotEqual(l1, l2)

    def test_data_lookup_eq(self):
        l1 = data_lookup(data=tf.constant("foo"))
        l2 = data_lookup(data=tf.constant("foo"))

        self.assertEqual(l1, l2)

    def test_data_lookup_not_eq(self):
        l1 = data_lookup(data=tf.constant("foo"))
        l2 = data_lookup(data=tf.constant("bar"))

        self.assertNotEqual(l1, l2)

    def test_optional_data_lookup_not_eq(self):
        l1 = data_lookup(data=tf.constant("foo"))
        l2 = data_lookup(data=None)

        self.assertNotEqual(l1, l2)

    def test_lookup_does_not_match_wrong_class(self):
        class WrongClass:
            def __init__(self):
                self.rank = 1
                self.distance = 0.1
                self.label = None
                self.embedding = None
                self.data = None

        l1 = basic_lookup()
        l2 = WrongClass()

        self.assertNotEqual(l1, l2)


if __name__ == "__main__":
    tf.test.main()



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Any, Type

from .nmslib_search import NMSLibSearch
from .search import Search

SEARCH_ALIASES: dict[str, Type[Search]] = {
    "NMSLibSearch": NMSLibSearch,
}


def make_search(config: dict[str, Any]) -> Search:
    """Creates a search instance from its config.

    This method is the reverse of `get_config`,
    capable of instantiating the same search from the config

    Args:
        config: A Python dictionary, typically the output of get_config.

    Returns:
        A search instance.
    """

    if config["canonical_name"] in SEARCH_ALIASES:
        search: Search = SEARCH_ALIASES[config["canonical_name"]](**config)
    else:
        raise ValueError(f"Unknown search type: {config['canonical_name']}")

    return search



import os
import re
import tempfile

import numpy as np
import pytest
import tensorflow as tf

from tensorflow_similarity.samplers import MultiShotFileSampler


def _create_random_image(filename, size=(32, 32)):
    filepath = os.path.join(tempfile.gettempdir(), filename)
    image = np.random.random(size + (3,)).astype(np.float32)
    tf.keras.utils.save_img(filepath, image)
    return filepath


@pytest.mark.parametrize("example_per_class", [2, 20])
def test_multi_shot_file_sampler(example_per_class):
    """Test MultiShotFileSampler with various sizes.

    Users may sample with replacement when creating batches, so check that we
    can handle when elements per class is either less than or greater than the
    total count of elements in the class.
    """
    filenames = ["1.jpg", "2.jpg", "3.jpg", "4.jpg", "5.jpg", "6.jpg"]
    filepaths = [_create_random_image(filename) for filename in filenames]
    images = [np.array(tf.keras.utils.load_img(path), dtype=np.float32) / 255 for path in filepaths]

    y = tf.constant([1, 2, 3, 1, 2, 3])
    x = tf.constant(filepaths)
    class_per_batch = 2
    batch_size = example_per_class * class_per_batch

    fs_sampler = MultiShotFileSampler(
        x=x,
        y=y,
        classes_per_batch=class_per_batch,
        examples_per_class_per_batch=example_per_class,
    )  # noqa

    batch_x, batch_y = fs_sampler.generate_batch(batch_id=606)

    assert len(batch_y) == batch_size
    assert len(batch_x) == batch_size
    num_classes, _ = tf.unique(batch_y)
    assert len(num_classes) == class_per_batch

    for x, y in zip(batch_x, batch_y):
        if y == 1:
            assert np.isclose(x.numpy(), images[0], atol=0.1).all() or np.isclose(x.numpy(), images[3], atol=0.1).all()
        elif y == 2:
            assert np.isclose(x.numpy(), images[1], atol=0.1).all() or np.isclose(x.numpy(), images[4], atol=0.1).all()
        elif y == 3:
            assert np.isclose(x.numpy(), images[2], atol=0.1).all() or np.isclose(x.numpy(), images[5], atol=0.1).all()


def test_msfs_get_slice():
    """Test the multi shot file sampler get_slice method."""
    filenames = ["1.jpg", "2.jpg", "3.jpg", "4.jpg"]
    filepaths = [_create_random_image(filename) for filename in filenames]
    images = [np.array(tf.keras.utils.load_img(path), dtype=np.float32) / 255 for path in filepaths]

    y = tf.constant(range(4))
    x = tf.constant(filepaths)

    fs_sampler = MultiShotFileSampler(x=x, y=y)
    # x and y are randomly shuffled so we fix the values here.
    fs_sampler._x = x
    fs_sampler._y = y
    slice_x, slice_y = fs_sampler.get_slice(1, 2)

    assert slice_x.shape == (2, 32, 32, 3)
    assert slice_y.shape == (2,)

    assert np.isclose(slice_x[0], images[1], atol=0.1).all()
    assert np.isclose(slice_x[1], images[2], atol=0.1).all()

    assert slice_y[0] == 1
    assert slice_y[1] == 2


def test_msms_properties():
    """Test the multi shot file sampler num_examples and shape"""
    filenames = ["1.jpg", "2.jpg", "3.jpg", "4.jpg"]
    filepaths = [_create_random_image(filename, (128, 96)) for filename in filenames]
    y = tf.constant(range(4))
    x = tf.constant(filepaths)

    fs_sampler = MultiShotFileSampler(x=x, y=y)

    assert fs_sampler.num_examples == 4
    assert fs_sampler.example_shape == (128, 96, 3)


def test_small_class_size(capsys):
    """Test examples_per_class is > the number of class examples."""
    filenames = ["1.jpg", "2.jpg", "3.jpg", "4.jpg"]
    filepaths = [_create_random_image(filename) for filename in filenames]

    y = tf.constant([1, 1, 1, 2])
    x = tf.constant(filepaths)

    fs_sampler = MultiShotFileSampler(x=x, y=y, classes_per_batch=2, examples_per_class_per_batch=3)

    _, batch_y = fs_sampler.generate_batch(0)

    y, _, class_counts = tf.unique_with_counts(batch_y)
    assert tf.math.reduce_all(tf.sort(y) == tf.constant([1, 2]))
    assert tf.math.reduce_all(class_counts == tf.constant([3, 3]))

    captured = capsys.readouterr()
    expected_msg = (
        "WARNING: Class 2 only has 1 unique examples, but "
        "examples_per_class is set to 3. The current batch will sample "
        "from class examples with replacement, but you may want to "
        "consider passing an Augmenter function or using the "
        "SingleShotMemorySampler()."
    )

    match = re.search(expected_msg, captured.out)
    assert bool(match)

    _, batch_y = fs_sampler.generate_batch(0)

    y, _, class_counts = tf.unique_with_counts(batch_y)
    assert tf.math.reduce_all(tf.sort(y) == tf.constant([1, 2]))
    assert tf.math.reduce_all(class_counts == tf.constant([3, 3]))

    # Subsequent batch should produce the sampler warning.
    captured = capsys.readouterr()
    match = re.search(expected_msg, captured.out)
    assert not bool(match)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Visualization utilities for TensorFlow Similarity."""
from __future__ import annotations

import itertools
from typing import Any

import numpy as np
import tensorflow as tf
from matplotlib import pyplot as plt

from tensorflow_similarity.types import FloatTensor, IntTensor


def confusion_matrix(
    y_pred: IntTensor,
    y_true: IntTensor,
    normalize: bool = True,
    labels: IntTensor | None = None,
    title: str = "Confusion matrix",
    cmap: str = "Blues",
    show: bool = True,
) -> tuple[Any, FloatTensor]:
    """Plot confusion matrix

    Args:
        y_pred: Model prediction returned by `model.match()`

        y_true: Expected class_id.

        normalize: Normalizes matrix values between 0 and 1.
        Defaults to True.

        labels: List of class string label to display instead of the class
        numerical ids. Defaults to None.

        title: Title of the confusion matrix. Defaults to 'Confusion matrix'.

        cmap: Color schema as CMAP. Defaults to 'Blues'.

        show: If the plot is going to be shown or not. Defaults to True.

    Returns:
        A Tuple containing the plot and confusion matrix.
    """

    with tf.device("/cpu:0"):
        # Ensure we are working with integer tensors.
        if not tf.is_tensor(y_pred):
            y_pred = tf.convert_to_tensor(np.array(y_pred))
        y_pred = tf.cast(y_pred, dtype="int32")
        if not tf.is_tensor(y_true):
            y_true = tf.convert_to_tensor(np.array(y_true))
        y_true = tf.cast(y_true, dtype="int32")

        cm = tf.math.confusion_matrix(y_true, y_pred)
        cm = tf.cast(cm, dtype=tf.keras.backend.floatx())
        accuracy = tf.linalg.trace(cm) / tf.math.reduce_sum(cm)
        misclass = 1 - accuracy

        if normalize:
            cm = tf.math.divide_no_nan(cm, tf.math.reduce_sum(cm, axis=1)[:, np.newaxis])

        f, ax = plt.subplots(figsize=(8, 6))
        im = ax.imshow(cm, interpolation="nearest", cmap=cmap)
        ax.set_title(title)
        f.colorbar(im)

        if labels is not None:
            tick_marks = np.arange(len(labels))
            ax.set_xticks(tick_marks)
            ax.set_xticklabels(labels, rotation=45)
            ax.set_yticks(tick_marks)
            ax.set_yticklabels(labels)

        cm_max = tf.math.reduce_max(cm)
        thresh = cm_max / 1.5 if normalize else cm_max / 2.0
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            val = cm[i, j]
            color = "white" if val > thresh else "black"
            txt = "%.2f" % val if val > 0.0 else "0"
            ax.text(j, i, txt, horizontalalignment="center", color=color)

        f.tight_layout()
        ax.set_ylabel("True label")
        ax.set_xlabel("Predicted label\naccuracy={:0.4f}; misclass={:0.4f}".format(accuracy, misclass))

        if show:
            plt.show()

        return ax, cm



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Triplet Loss.
    FaceNet: A Unified Embedding for Face Recognition and Clustering
    https://arxiv.org/abs/1503.03832
"""
from __future__ import annotations

from typing import Any

import tensorflow as tf

from tensorflow_similarity.algebra import build_masks
from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.types import FloatTensor, IntTensor

from .metric_loss import MetricLoss
from .utils import compute_loss, negative_distances, positive_distances


def triplet_loss(
    query_labels: IntTensor,
    query_embeddings: FloatTensor,
    key_labels: IntTensor,
    key_embeddings: FloatTensor,
    distance: Distance,
    remove_diagonal: bool = True,
    positive_mining_strategy: str = "hard",
    negative_mining_strategy: str = "semi-hard",
    margin: float | None = None,
) -> Any:
    """Triplet loss computations.

    Args:
        query_labels: labels associated with the query embed.
        query_embeddings: Embedded query examples.
        key_labels: labels associated with the key embed.
        key_embeddings: Embedded key examples.
        distance: Which distance function to use to compute the pairwise
            distances between embeddings.
        remove_diagonal: If True, set the diagonal to False in positive pair
            mask.
        positive_mining_strategy: What mining strategy to use to select
            embedding from the same class. {'easy', 'hard'}
        negative_mining_strategy: What mining strategy to use for select the
            embedding from the different class. {'hard', 'semi-hard', 'easy'}
        margin: Defines the target margin between positive and negative pairs,
            e.g., a margin of 1.0 means that the positive and negative
            distances should be 1.0 apart. If None, then a soft margin is used.
            A soft margin can be beneficial to pull together samples from the
            same class as much as possible. See the paper for more details
            https://arxiv.org/pdf/1703.07737.pdf. Defaults to None.

    Returns:
        Loss: The loss value for the current batch.
    """

    # [Label]
    # ! Weirdness to be investigated
    # do not remove this code. It is actually needed for specific situation
    # Reshape label tensor to [batch_size, 1] if not already in that format.
    # labels = tf.reshape(labels, (labels.shape[0], 1))
    batch_size = tf.size(query_labels)

    # [distances]
    pairwise_distances = distance(query_embeddings, key_embeddings)

    # [masks]
    positive_mask, negative_mask = build_masks(
        query_labels,
        key_labels,
        batch_size=batch_size,
        remove_diagonal=remove_diagonal,
    )

    # [Positive distance computation]
    pos_distances, pos_idxs = positive_distances(
        positive_mining_strategy,
        pairwise_distances,
        positive_mask,
    )

    # [Negative distances computation]
    neg_distances, neg_idxs = negative_distances(
        negative_mining_strategy,
        pairwise_distances,
        negative_mask,
        positive_mask,
    )

    # [Triplet loss computation]
    triplet_loss = compute_loss(pos_distances, neg_distances, margin)

    return triplet_loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class TripletLoss(MetricLoss):
    """Computes the triplet loss in an online fashion.

    This loss encourages the positive distances between a pair of embeddings
    with the same labels to be smaller than the minimum negative distances
    between pair of embeddings of different labels.
    See: https://arxiv.org/abs/1503.03832 for the original paper.


    `y_true` must be  a 1-D integer `Tensor` of shape (batch_size,).
    It's values represent the classes associated with the examples as
    **integer  values**.

    `y_pred` must be 2-D float `Tensor`  of L2 normalized embedding vectors.
    you can use the layer `tensorflow_similarity.layers.L2Embedding()` as the
    last layer of your model to ensure your model output is properly
    normalized.
    """

    def __init__(
        self,
        distance: Distance | str = "cosine",
        positive_mining_strategy: str = "hard",
        negative_mining_strategy: str = "semi-hard",
        margin: float | None = None,
        name: str = "TripletLoss",
        **kwargs,
    ):
        """Initializes the TripletLoss.

        Args:
            distance: Which distance function to use to compute the pairwise
                distances between embeddings.
            remove_diagonal: If True, set the diagonal to False in positive pair
                mask.
            positive_mining_strategy: What mining strategy to use to select
                embedding from the same class. {'easy', 'hard'}
            negative_mining_strategy: What mining strategy to use for select the
                embedding from the different class. {'hard', 'semi-hard', 'easy'}
            margin: Defines the target margin between positive and negative pairs,
                e.g., a margin of 1.0 means that the positive and negative
                distances should be 1.0 apart. If None, then a soft margin is used.
                A soft margin can be beneficial to pull together samples from the
                same class as much as possible. See the paper for more details
                https://arxiv.org/pdf/1703.07737.pdf. Defaults to None.
            name: Loss name. Defaults to "TripletLoss".

        Raises:
            ValueError: Invalid positive mining strategy.
            ValueError: Invalid negative mining strategy.
        """

        # distance canonicalization
        distance = distance_canonicalizer(distance)
        self.distance = distance

        # sanity checks
        if positive_mining_strategy not in ["easy", "hard"]:
            raise ValueError("Invalid positive mining strategy")

        if negative_mining_strategy not in ["easy", "hard", "semi-hard"]:
            raise ValueError("Invalid negative mining strategy")

        super().__init__(
            triplet_loss,
            name=name,
            distance=distance,
            positive_mining_strategy=positive_mining_strategy,
            negative_mining_strategy=negative_mining_strategy,
            margin=margin,
            **kwargs,
        )



import tensorflow as tf
from absl.testing import parameterized
from tensorflow.keras.losses import Reduction
from tensorflow.python.framework import combinations

from tensorflow_similarity import losses

from . import utils


@combinations.generate(combinations.combine(mode=["graph", "eager"]))
class TripletLossTest(tf.test.TestCase, parameterized.TestCase):
    def test_config(self):
        tpl_obj = losses.TripletLoss(reduction=Reduction.SUM, name="triplet_loss", distance="cosine")
        self.assertEqual(tpl_obj.distance.name, "cosine")
        self.assertEqual(tpl_obj.name, "triplet_loss")
        self.assertEqual(tpl_obj.reduction, Reduction.SUM)

    @parameterized.named_parameters(
        {"testcase_name": "_soft_margin", "margin": None, "expected_loss": 0.31326169},
        {"testcase_name": "_fixed_margin", "margin": 1.1, "expected_loss": 0.1},
    )
    def test_all_correct_unweighted(self, margin, expected_loss):
        """Test that assumes a perfect embedding.

        Using cosine distance, the prefect embedding will contain one hot encoded
        embeddings with a value of 1.0 in the class position and 0 everywhere else.

        Using Soft Margin we expect `pos_dist - neg_dist = -1.0` yielding
        `ln(1 + e^(-1.0)) = 0.31326169`.

        Using a fixed margin of 1.1 we expect the loss to stop at 0.1.
        """
        tpl_obj = losses.TripletLoss(reduction=Reduction.SUM_OVER_BATCH_SIZE, margin=margin)
        y_true, y_preds = utils.generate_perfect_test_batch()
        loss = tpl_obj(y_true, y_preds)
        self.assertAlmostEqual(self.evaluate(loss), expected_loss, 3)

    @parameterized.named_parameters(
        {"testcase_name": "_soft_margin", "margin": None, "expected_loss": 0.69314718},
        {"testcase_name": "_fixed_margin", "margin": 1.0, "expected_loss": 1.0},
    )
    def test_all_mismatch_unweighted(self, margin, expected_loss):
        """Test that assumes none of the classes match.

        Using cosine distance, each embedding is one hot encoded with a value of
        1.0 in the class position and 0 everywhere else.

        Using Soft Margin we expect `pos_dist - neg_dist = 0.0` yielding
        `ln(1 + e^(0.0)) = 0.69314718`.

        Using a fixed margin of 1.0 we expect the loss to stop at 1.0.
        """
        tpl_obj = losses.TripletLoss(reduction=Reduction.SUM_OVER_BATCH_SIZE, margin=margin)
        y_true, y_preds = utils.generate_bad_test_batch()
        loss = tpl_obj(y_true, y_preds)
        self.assertAlmostEqual(self.evaluate(loss), expected_loss, 3)

    @parameterized.named_parameters(
        {"testcase_name": "_soft_margin", "margin": None, "expected_loss": 0.31326169},
        {"testcase_name": "_fixed_margin", "margin": 1.1, "expected_loss": 0.1},
    )
    def test_no_reduction(self, margin, expected_loss):
        tpl_obj = losses.TripletLoss(reduction=Reduction.NONE, margin=margin)
        y_true, y_preds = utils.generate_perfect_test_batch()
        loss = tpl_obj(y_true, y_preds)
        loss = self.evaluate(loss)
        expected_loss = self.evaluate(tf.fill(y_true.shape, expected_loss))
        self.assertArrayNear(loss, expected_loss, 0.001)

    @parameterized.named_parameters(
        {"testcase_name": "_soft_margin", "margin": None, "expected_loss": 0.31326169},
        {"testcase_name": "_fixed_margin", "margin": 1.1, "expected_loss": 0.1},
    )
    def test_sum_reduction(self, margin, expected_loss):
        tpl_obj = losses.TripletLoss(reduction=Reduction.SUM, margin=margin)
        y_true, y_preds = utils.generate_perfect_test_batch()
        loss = tpl_obj(y_true, y_preds)
        expected_loss = y_true.shape[0] * expected_loss
        self.assertAlmostEqual(self.evaluate(loss), expected_loss, 3)

    @parameterized.named_parameters(
        {
            "testcase_name": "_easy",
            "positive_mining_strategy": "easy",
            "negative_mining_strategy": "easy",
            "expected_loss": 0.333,
        },
        {
            "testcase_name": "_hard",
            "positive_mining_strategy": "hard",
            "negative_mining_strategy": "hard",
            "expected_loss": 2.0,
        },
    )
    def test_mining(self, positive_mining_strategy, negative_mining_strategy, expected_loss):
        """Test the easy mining strategy

        Easy mining strategy will always pick the minimum distance for the
        positive or negative pair. This means that the pair for positive
        should always be [1.0, 0.0] and for negative should always be [0.0, 1.0].
        This means that the loss should be
        sum([0.0, 0.0, 1.0, 0.0, 0.0, 1.0]) / 6 = 0.333.

        Hard ming strategy will always pick the maximum distance for the
        positive or negative pair. This means that the pair for positive
        should always be [0.0, 1.0] and for negative should always be [1.0, 0.0].
        This means that the loss should be
        sum([1.0, 1.0, 1.0, 1.0, 1.0, 1.0]) / 6 = 2.0.
        """
        tpl_obj = losses.TripletLoss(
            reduction=Reduction.SUM_OVER_BATCH_SIZE,
            positive_mining_strategy=positive_mining_strategy,
            negative_mining_strategy=negative_mining_strategy,
            margin=1.0,
        )
        labels = tf.expand_dims(tf.constant([0, 0, 0, 1, 1, 1], dtype=(tf.int32)), axis=1)
        embeddings = tf.constant(
            [[1.0, 0.0], [1.0, 0.0], [0.0, 1.0], [0.0, 1.0], [0.0, 1.0], [1.0, 0.0]], dtype=(tf.float32)
        )
        loss = tpl_obj(labels, embeddings)
        self.assertAlmostEqual(self.evaluate(loss), expected_loss, 3)

    def test_semi_hard_mining(self):
        """Test the easy positive mining and semi-hard negative mining strategy.

        Easy positive mining strategy will just pick the other positive example,
        i.e., the pos_idxs will be [1,0,3,2]

        Semi-hard negative mining strategy will pick the closest negative example
        that is > the max distance between the positive pairs, if one exists,
        otherwise it will pick the maximal negative example. The neg_idxs will be
        [3,2,1,1] where the second positive example actully has no negatives that
        are > the max distance between the maximal positive.This is why the loss
        for the second positive example is > 1.0.

        The expected pairwise distances are:
            [[0.0,  0.5,  1.0,     0.98],   # Choose 0.98 because it's > 0.5 but < 1.0
             [0.5,  0.0,  0.13,    0.12],   # Choose 0.13 because it's there is no > 0.5
             [1.0,  0.13, 0.0,     0.00005],# Choose 0.13 because it's > 5e-5 but < 1.0
             [0.98, 0.12, 0.00005, 0.0]]    # Choose 0.12 because it's > 5e-5 but < 0.98
        """
        tpl_obj = losses.TripletLoss(
            reduction=Reduction.NONE,
            positive_mining_strategy="easy",
            negative_mining_strategy="semi-hard",
            margin=1.0,
        )
        labels = tf.expand_dims(tf.constant([0, 0, 1, 1], dtype=(tf.int32)), axis=1)
        embeddings = tf.constant(
            [[1.0, 0.0], [0.5, 0.8660254], [0.0, 1.0], [0.01010049, 0.999949]], dtype=(tf.float32)
        )
        loss = tpl_obj(labels, embeddings)
        expected_loss = tf.constant([[0.51], [1.366], [0.866], [0.871]], dtype=(tf.float32))
        self.assertArrayNear(self.evaluate(loss), self.evaluate(expected_loss), 0.001)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import nullcontext

import tensorflow as tf
from absl.testing import parameterized

from tensorflow_similarity.layers import ActivationStdLoggingLayer
from tensorflow_similarity.losses import SimSiamLoss
from tensorflow_similarity.models import ContrastiveModel, create_contrastive_model


class ContrastiveModelTest(tf.test.TestCase, parameterized.TestCase):
    @parameterized.named_parameters(
        {
            "testcase_name": "_no_strategy",
            "strategy": None,
        },
        {
            "testcase_name": "_mirrored_strategy",
            "strategy": lambda: tf.distribute.MirroredStrategy().scope(),
        },
    )
    def test_save_and_reload(self, strategy):
        """Test save and load of ContrastiveModel.
        Testing it also in a MirroredStrategy on GPU if available, to check fix for
        issue #287.
        """
        out_dir = self.get_temp_dir()

        with strategy() if strategy is not None else nullcontext():
            backbone_input = tf.keras.layers.Input(shape=(3,))
            backbone_output = tf.keras.layers.Dense(4)(backbone_input)
            backbone = tf.keras.Model(
                inputs=backbone_input,
                outputs=backbone_output,
            )

            model = create_contrastive_model(
                backbone=backbone,
                algorithm="simsiam",
            )
            opt = tf.keras.optimizers.RMSprop(learning_rate=0.5)

            model.compile(optimizer=opt, loss=SimSiamLoss())

        # test data
        x = tf.constant([[1, 1, 3], [3, 1, 2]], dtype="int64")

        # create dataset with two views
        ds = tf.data.Dataset.from_tensors(x)
        ds = ds.map(lambda x: (x, x), num_parallel_calls=tf.data.AUTOTUNE)

        model.fit(ds)

        # save
        model.save(out_dir)

        # reload and test loaded model
        with strategy() if strategy is not None else nullcontext():
            loaded_model = tf.keras.models.load_model(
                out_dir,
                custom_objects={
                    "ContrastiveModel": ContrastiveModel,
                    "ActivationStdLoggingLayer": ActivationStdLoggingLayer,
                },
            )

        pred = loaded_model.predict(x)

        self.assertEqual(loaded_model.algorithm, "simsiam")
        self.assertEqual(loaded_model.optimizer.lr, 0.5)
        self.assertAllEqual(loaded_model.backbone.input_shape, (None, 3))
        self.assertAllEqual(loaded_model.backbone.output_shape, (None, 4))
        self.assertAllEqual(loaded_model.projector.input_shape, (None, 4))
        self.assertAllEqual(loaded_model.projector.output_shape, (None, 512))
        self.assertAllEqual(loaded_model.predictor.input_shape, (None, 512))
        self.assertAllEqual(loaded_model.predictor.output_shape, (None, 512))
        self.assertAllEqual(pred.shape, (2, 512))
        self.assertAllEqual(model.predict(x), pred)


if __name__ == "__main__":
    tf.test.main()



import tensorflow as tf

from tensorflow_similarity.samplers import tfdata_sampler as tfds


class TestFilterClasses(tf.test.TestCase):
    def setUp(self):
        self.ds = tf.data.Dataset.from_tensor_slices(
            (
                tf.constant([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0]),
                tf.constant([1, 1, 2, 2, 4, 3, 4, 3]),
            )
        )
        self.classes = {c: 0 for c in range(1, 5)}

    def test_filter_classes(self):
        ds = tfds.filter_classes(self.ds, [2, 4])
        for _, label in ds:
            self.classes[label.numpy()] += 1
        self.assertEqual(self.classes[1], 0)
        self.assertEqual(self.classes[2], 2)
        self.assertEqual(self.classes[3], 0)
        self.assertEqual(self.classes[4], 2)

    def test_filter_empty_class_list(self):
        ds = tfds.filter_classes(self.ds)
        for _, label in ds:
            self.classes[label.numpy()] += 1
        self.assertEqual(self.classes[1], 2)
        self.assertEqual(self.classes[2], 2)
        self.assertEqual(self.classes[3], 2)
        self.assertEqual(self.classes[4], 2)


class TestCreateGroupedDataset(tf.test.TestCase):
    def setUp(self):
        self.ds = tf.data.Dataset.from_tensor_slices(
            (
                tf.constant([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0]),
                tf.constant([1, 1, 2, 2, 3, 3, 4, 4]),
            )
        )
        self.window_size = self.ds.cardinality().numpy()

    def test_returns_correct_number_of_datasets(self):
        cid_datasets = tfds.create_grouped_dataset(self.ds, self.window_size)
        self.assertLen(cid_datasets, 4)

    def test_returns_correct_number_of_datasets_with_total_examples(self):
        cid_datasets = tfds.create_grouped_dataset(self.ds, self.window_size, total_examples=1)
        self.assertLen(cid_datasets, 4)

        # test that each cid dataset has only 1 example that is repeated.
        for elem in cid_datasets:
            self.assertAllEqual(
                list(elem.take(1).as_numpy_iterator()),
                list(elem.take(1).as_numpy_iterator()),
            )

    def test_returns_correct_number_of_datasets_with_buffer_size(self):
        cid_datasets = tfds.create_grouped_dataset(self.ds, self.window_size, buffer_size=2)
        self.assertEqual(len(cid_datasets), 4)

    def test_datasets_repeat(self):
        cid_datasets = tfds.create_grouped_dataset(self.ds, self.window_size)
        for cid_ds in cid_datasets:
            self.assertTrue(cid_ds.element_spec, self.ds.element_spec)

            # Check that repeating groups of 2 elements all equal each other.
            elements = list(cid_ds.take(6).as_numpy_iterator())
            self.assertAllEqual(elements[:2], elements[2:4])
            self.assertAllEqual(elements[:2], elements[4:])

    def test_datasets_shuffled(self):
        cid_datasets = tfds.create_grouped_dataset(self.ds, self.window_size, buffer_size=4)
        # check that including the buffer shuffles the values in each cid_ds.
        for cid_ds in cid_datasets:
            self.assertNotEqual(
                list(cid_ds.take(20).as_numpy_iterator()),
                list(cid_ds.take(20).as_numpy_iterator()),
            )


class TestCreateChoicesDataset(tf.test.TestCase):
    def setUp(self):
        self.classes = {c: 0 for c in range(5)}

    def test_choice_values_with_repetition(self):
        # Test that each class appears exactly examples_per_class * repeat times
        num_classes = 5
        examples_per_class = 2
        num_repeats = 2
        dataset = tfds.create_choices_dataset(num_classes, examples_per_class)
        for label in dataset.take(num_classes * examples_per_class * num_repeats):
            self.classes[label.numpy()] += 1

        self.assertSetEqual(set(self.classes.keys()), set(range(5)))
        self.assertEqual(self.classes[0], 4)
        self.assertEqual(self.classes[1], 4)
        self.assertEqual(self.classes[2], 4)
        self.assertEqual(self.classes[3], 4)
        self.assertEqual(self.classes[4], 4)


def dummy_augmenter(x):
    return x + 10


class TestAugmenter(tf.test.TestCase):
    def setUp(self):
        self.ds = tf.data.Dataset.range(1, 5)

    def test_apply_augmentation_no_warmup(self):
        augmented_ds = tfds.apply_augmenter_ds(self.ds, dummy_augmenter)

        for x in augmented_ds.batch(2).as_numpy_iterator():
            self.assertAllEqual(x, [11, 12])
            break

    def test_apply_augmentation_with_warmup(self):
        warmup = 2
        batch_size = 2
        augmented_ds = tfds.apply_augmenter_ds(self.ds, dummy_augmenter, warmup)

        for i, x in enumerate(augmented_ds.batch(batch_size).take(2)):
            if i < warmup // batch_size:
                self.assertAllEqual(x, tf.constant([1, 2]))
            else:
                self.assertAllEqual(x, tf.constant([13, 14]))

    def test_apply_augmentation_with_warmup_and_repeats(self):
        values = {c: 0 for c in [1, 2, 3, 4, 11, 12, 13, 14]}
        warmup = 2
        augmented_ds = tfds.apply_augmenter_ds(self.ds.repeat(2), dummy_augmenter, warmup)

        for x in augmented_ds:
            values[x.numpy()] += 1
        self.assertEqual(values[1], 1)
        self.assertEqual(values[2], 1)
        self.assertEqual(values[3], 0)
        self.assertEqual(values[4], 0)
        self.assertEqual(values[11], 1)
        self.assertEqual(values[12], 1)
        self.assertEqual(values[13], 2)
        self.assertEqual(values[14], 2)


class TestTFDataSampler(tf.test.TestCase):
    def setUp(self):
        self.ds = tf.data.Dataset.from_tensor_slices(
            (
                tf.expand_dims(tf.constant([0, 1, 2, 3, 4, 5], dtype=tf.int32), axis=-1),
                tf.constant([1, 1, 1, 2, 2, 2], dtype=tf.int32),
            )
        )
        self.expected_element_spec = (
            tf.TensorSpec(shape=(None, 1), dtype=tf.int32),
            tf.TensorSpec(shape=(None,), dtype=tf.int32),
        )

    def test_cardinality_is_finite(self):
        # Test that an exception is raised when the input dataset is infinite
        ds = tf.data.Dataset.from_tensors([1]).repeat()
        with self.assertRaisesWithLiteralMatch(ValueError, "Dataset must be finite."):
            tfds.TFDataSampler(ds)

    def test_cardinality_is_known(self):
        # Test that an exception is raised when the input dataset has unknown cardinality
        ds = tf.data.Dataset.from_tensor_slices([1, 2, 3]).shuffle(3).filter(lambda x: x > 1)
        with self.assertRaisesWithLiteralMatch(ValueError, "Dataset cardinality must be known."):
            tfds.TFDataSampler(ds)

    def test_output_batch_size(self):
        # Test that the output batch size is correct
        out_ds = tfds.TFDataSampler(self.ds, classes_per_batch=2, examples_per_class_per_batch=2)
        self.assertEqual(out_ds.element_spec, self.expected_element_spec)
        for _, y in out_ds.take(1):
            self.assertLen(y, 4)

    def test_output_classes_per_batch(self):
        # Test that the number of classes per batch is correct
        out_ds = tfds.TFDataSampler(self.ds, classes_per_batch=1)
        self.assertEqual(out_ds.element_spec, self.expected_element_spec)
        for _, y in out_ds.take(2):
            self.assertLen(tf.unique(y).y, 1)

    def test_output_examples_per_class_per_batch(self):
        # Test that the number of examples per class per batch is correct
        out_ds = tfds.TFDataSampler(self.ds, classes_per_batch=2, examples_per_class_per_batch=3)
        self.assertEqual(out_ds.element_spec, self.expected_element_spec)
        for _, y in out_ds.take(2):
            self.assertAllEqual(tf.math.bincount(y), [0, 3, 3])

    def test_output_class_list(self):
        # Test that the class list is correctly used
        out_ds = tfds.TFDataSampler(self.ds, class_list=[1])
        self.assertEqual(out_ds.element_spec, self.expected_element_spec)
        for _, y in out_ds.take(4):
            self.assertAllEqual(tf.unique(y).y, [1])

    def test_output_total_examples_per_class(self):
        # Test that the total number of examples per class is correctly used
        out_ds = tfds.TFDataSampler(self.ds, classes_per_batch=2, total_examples_per_class=2)
        self.assertEqual(out_ds.element_spec, self.expected_element_spec)
        for x, y in out_ds.take(2):
            for x_, y_ in zip(x, y):
                if y_ == 1:
                    self.assertLessEqual(x_, 2)
                elif y_ == 2:
                    self.assertGreater(x_, 2)
                else:
                    raise ValueError("Unexpected class")

    def test_output_augmenter_with_warmup_and_repeat(self):
        # Test that the augmenter is correctly applied
        out_ds = tfds.TFDataSampler(
            self.ds,
            classes_per_batch=2,
            examples_per_class_per_batch=3,
            augmenter=lambda x, y: (x + 6, y),
            warmup=6,
        )
        self.assertEqual(out_ds.element_spec, self.expected_element_spec)
        values = {c: 0 for c in range(12)}
        for x, y in out_ds.take(3):
            for x_ in x:
                values[x_.numpy()[0]] += 1
        self.assertEqual(values[1], 1)
        self.assertEqual(values[2], 1)
        self.assertEqual(values[3], 1)
        self.assertEqual(values[4], 1)
        self.assertEqual(values[5], 1)
        self.assertEqual(values[6], 2)
        self.assertEqual(values[7], 2)
        self.assertEqual(values[8], 2)
        self.assertEqual(values[9], 2)
        self.assertEqual(values[10], 2)
        self.assertEqual(values[11], 2)

    def test_output_load_fn(self):
        out_ds = tfds.TFDataSampler(
            self.ds,
            classes_per_batch=2,
            total_examples_per_class=2,
            load_fn=lambda x, y, *args: ("loaded_img", y, *args),
        )
        for x, y in out_ds:
            self.assertAllEqual(x.numpy(), tf.constant([b"loaded_img"] * 4))
            break

    def test_label_output_dict(self):
        ds = tf.data.Dataset.from_tensor_slices(
            (
                tf.constant([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0]),
                {
                    "A": tf.constant([1, 1, 1, 1, 1, 1, 1, 1]),
                    "B": tf.constant([3, 3, 3, 3, 2, 2, 2, 2]),
                },
            )
        )
        out_ds = tfds.TFDataSampler(
            ds,
            classes_per_batch=2,
            examples_per_class_per_batch=3,
            label_output="B",
        )
        for _, y in out_ds.take(4):
            self.assertSetEqual(set(tf.unique(y["B"]).y.numpy()), set([2, 3]))
            self.assertAllEqual(tf.math.bincount(y["B"]), [0, 0, 3, 3])

    def test_label_output_tuple(self):
        ds = tf.data.Dataset.from_tensor_slices(
            (
                tf.constant([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0]),
                (
                    tf.constant([1, 1, 1, 1, 1, 1, 1, 1]),
                    tf.constant([3, 3, 3, 3, 2, 2, 2, 2]),
                ),
            )
        )
        out_ds = tfds.TFDataSampler(
            ds,
            classes_per_batch=2,
            examples_per_class_per_batch=3,
            label_output=1,
        )
        for _, y in out_ds.take(4):
            self.assertSetEqual(set(tf.unique(y[1]).y.numpy()), set([2, 3]))
            self.assertAllEqual(tf.math.bincount(y[1]), [0, 0, 3, 3])


if __name__ == "__main__":
    tf.test.main()



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import annotations

import tensorflow as tf

from tensorflow_similarity.augmenters.augmentation_utils.random_apply import (
    random_apply,
)
from tensorflow_similarity.types import Tensor


def random_solarize(
    image: Tensor,
    p: float = 0.2,
    pixel_min: int = 0,
    pixel_max: int = 255,
    thresh: int = 10,
) -> Tensor:
    def _transform(image: Tensor) -> Tensor:
        return tf.where(image < thresh, image, pixel_max - image)

    return random_apply(_transform, p=p, x=image)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from abc import ABC, abstractmethod

from tensorflow_similarity.types import FloatTensor


class ClassificationMetric(ABC):
    """Abstract base class for computing classification metrics.

    Args:
        name: Name associated with a specific metric object, e.g.,
        accuracy@0.1

        canonical_name: The canonical name associated with metric, e.g.,
        accuracy

        maximize: If True, we attempt to maximize the calibration metric
        as the distance increases.

        increasing: If True, the metric should increase as the distance
        increases.

    `ClassificationMetric` measure the matching classification between the
    query label and the label derived from the set of lookup results.

    The `compute()` method supports computing the metric for a set of values,
    where each value represents the counts at a specific distance threshold.
    """

    def __init__(
        self, name: str = "", canonical_name: str = "", maximize: bool = True, increasing: bool = True
    ) -> None:
        self.name = name
        self.canonical_name = canonical_name
        self.maximize = maximize
        self.increasing = increasing

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return "%s:%s" % (self.canonical_name, self.name)

    def get_config(self):
        return {
            "name": self.name,
            "canonical_name": self.canonical_name,
            "maximize": self.maximize,
            "increasing": self.increasing,
        }

    @abstractmethod
    def compute(self, tp: FloatTensor, fp: FloatTensor, tn: FloatTensor, fn: FloatTensor, count: int) -> FloatTensor:
        """Compute the classification metric.

        Args:
            tp: A 1D FloatTensor containing the count of True Positives at each
            distance threshold.

            fp: A 1D FloatTensor containing the count of False Positives at
            each distance threshold.

            tn: A 1D FloatTensor containing the count of True Negatives at each
            distance threshold.

            fn: A 1D FloatTensor containing the count of False Negatives at
            each distance threshold.

            count: The total number of queries

        Returns:
            A 1D FloatTensor containing the metric at each distance threshold.
        """



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from tensorflow_similarity.losses import TripletLoss
from tensorflow_similarity.models import SimilarityModel


class SimilarityModelTest(tf.test.TestCase):
    # TODO(ovallis): Add tests for graph mode.
    def test_save_and_reload(self):
        out_dir = self.get_temp_dir()

        inputs = tf.keras.layers.Input(shape=(3,))
        outputs = tf.keras.layers.Dense(2)(inputs)
        model = SimilarityModel(inputs, outputs)
        model.compile(optimizer="adam", loss=TripletLoss())

        # index data
        x = tf.constant([[1, 1, 3], [3, 1, 2]], dtype="float32")
        y = tf.constant([1, 2])
        model.index(x, y)

        # save
        model.save(out_dir)

        # reload
        loaded_model = tf.keras.models.load_model(out_dir)
        loaded_model.load_index(out_dir)
        self.assertEqual(loaded_model._index.size(), len(y))

    def test_save_no_compile(self):
        out_dir = self.get_temp_dir()

        inputs = tf.keras.layers.Input(shape=(3,))
        outputs = tf.keras.layers.Dense(2)(inputs)
        model = SimilarityModel(inputs, outputs)

        model.save(out_dir)
        model2 = tf.keras.models.load_model(out_dir)
        self.assertIsInstance(model2, type(model))

    def test_index_single(self):
        """Unit Test for issues #161 & #162"""
        inputs = tf.keras.layers.Input(shape=(3,))
        outputs = tf.keras.layers.Dense(2)(inputs)
        model = SimilarityModel(inputs, outputs)
        model.compile(optimizer="adam", loss=TripletLoss())

        # index data
        x = tf.constant([1, 1, 3], dtype="float32")
        y = tf.constant([1])

        # run individual sample & index
        model.index_single(x, y, data=x)
        self.assertEqual(model._index.size(), 1)


if __name__ == "__main__":
    tf.test.main()
